"""NDG Security Obligation type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: obligation.py 7103 2010-06-25 21:03:35Z pjkersha $"
from ndg.xacml.utils import TypedList
from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.attributevalue import AttributeValue


class AttributeAssignment(AttributeValue):
    """XACML AttributeAssignment type"""
       

class Effect(object):
    """Define effect type for Obligation
    
    @cvar PERMIT_STR: permit decision string
    @type PERMIT_STR: string
    
    @cvar DENY_STR: deny decision string
    @type DENY_STR: string
    
    @cvar TYPES: list of valid effect strings
    @type TYPES: tuple
    
    @ivar __value: obligation effect
    @type value: None / basestring
    """
    
    # "Permit" effect type
    PERMIT_STR = "Permit"
    
    # "Deny" effect type
    DENY_STR = "Deny"
        
    TYPES = (PERMIT_STR, DENY_STR)
    
    __slots__ = ('__value',)
    
    def __init__(self, effectType):
        """Initialise attributes giving an effect type
        @param effectType: 
        @type effectType: 
        """
        self.__value = None
        self.value = effectType

    def __getstate__(self):
        '''Enable pickling
        
        @return: instance attributes dictionary
        @rtype: dict
        '''
        _dict = {}
        for attrName in Effect.__slots__:
            # Ugly hack to allow for derived classes setting private member
            # variables
            if attrName.startswith('__'):
                attrName = "_Effect" + attrName
                
            _dict[attrName] = getattr(self, attrName)
            
        return _dict
  
    def __setstate__(self, attrDict):
        '''Enable pickling
        @param attrDict: instance attributes dictionary
        @type attrDict: dict
        '''
        for attrName, val in attrDict.items():
            setattr(self, attrName, val)
            
    def _setValue(self, value):
        """Set effect
        
        @param value: effect
        @type value: ndg.xacml.core.obligation.Effect or basestring
        @raise TypeError: incorrect type for input
        @raise AttributeError: effect value not recognised
        """
        if isinstance(value, Effect):
            # Cast to string
            value = str(value)
            
        elif not isinstance(value, basestring):
            raise TypeError('Expecting string or Effect instance for '
                            '"value" attribute; got %r instead' % type(value))
            
        if value not in self.__class__.TYPES:
            raise AttributeError('Permissable effect types are %r; got '
                                 '%r instead' % (Effect.TYPES, value))
        self.__value = value
        
    def _getValue(self):
        """Get effect
        
        @return: effect value
        @rtype: ndg.xacml.core.obligation.Effect
        """
        return self.__value
    
    value = property(fget=_getValue, fset=_setValue, doc="Effect value")
    
    def __str__(self):
        """
        @return: effect as a string
        @rtype: basestring
        """
        return self.__value

    def __eq__(self, effect):
        """
        @param effect: ndg.xacml.core.obligation.Effect or basestring
        @raise TypeError: incorrect type for input
        @raise AttributeError: effect value not recognised
        """
        if isinstance(effect, Effect):
            # Cast to string
            value = effect.value
            
        elif isinstance(effect, basestring):
            value = effect
            
        else:
            raise TypeError('Expecting string or Effect instance for '
                            'input effect value; got %r instead' % type(value))
            
        if value not in self.__class__.TYPES:
            raise AttributeError('Permissable effect types are %r; got '
                                 '%r instead' % (Effect.TYPES, value))
            
        return self.__value == value
    

class Obligation(XacmlCoreBase):
    """XACML Obligation type
    
    @cvar ELEMENT_LOCAL_NAME: XML element local name
    @type ELEMENT_LOCAL_NAME: string
    @cvar ATTRIBUTE_ASSIGNMENTS_ELEMENT_LOCAL_NAME: attribute assignments XML 
    attribute name
    @type ATTRIBUTE_ASSIGNMENTS_ELEMENT_LOCAL_NAME: string
    @cvar OBLIGATION_ID_ATTRIB_NAME: obligation ID XML attribute name
    @type OBLIGATION_ID_ATTRIB_NAME: string
    
    @ivar __attributeAssignments: list of attribute assignments
    @type __attributeAssignments: ndg.xacml.utils.TypedList
    @ivar __obligationId: obligation id
    @type __obligationId: NoneType / basestring
    @ivar __fulfillOn: fulfil on condition
    @type __fulfillOn: NoneType / basestring
    """
    ELEMENT_LOCAL_NAME = 'Obligation'
    ATTRIBUTE_ASSIGNMENTS_ELEMENT_LOCAL_NAME = 'AttributeAssignment'
    FULFILLON_ELEMENT_LOCAL_NAME = 'FulfillOn'
    OBLIGATION_ID_ATTRIB_NAME = 'ObligationId'
    
    __slots__ = ('__attributeAssignments', '__obligationId', '__fulfillOn')
    
    def __init__(self):
        """Initialise attributes"""
        self.__attributeAssignments = TypedList(AttributeAssignment)
        self.__obligationId = None
        self.__fulfillOn = None
        
    @property
    def attributeAssignments(self):
        """Obligation attribute assignments
        @return: attribute assignments
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__attributeAssignments
    
    @property
    def obligationId(self):
        """obligation Id
        @return: obligation id
        @rtype: NoneType / basestring
        """
        return self.__obligationId

    @obligationId.setter
    def obligationId(self, value):
        """obligation Id
        
        @param value: obligation id
        @type value: NoneType / basestring
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "obligationId" attribute; '
                            'got %r' % (basestring, type(value)))
            
        self.__obligationId = value        
    
    @property
    def fulfillOn(self):
        """Fulfill obligation on the given effect Permit/Deny
        
        @return: fulfil condition
        @rtype: NoneType / basestring
        """
        return self.__fulfillOn

    @fulfillOn.setter
    def fulfillOn(self, value):
        """Fulfill obligation on the given effect Permit/Deny
        @param value: fulfil condition
        @type value: NoneType / basestring
        @raise TypeError: incorrect type for input
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "fulfillOn" attribute; got '
                            '%r' % (Effect, type(value)))
            
        self.__fulfillOn = value
        



"""NDG Security Policy Defaults type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "19/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: policydefaults.py 7108 2010-06-28 09:19:26Z pjkersha $"
from ndg.xacml.core import XacmlCoreBase


class PolicyDefaults(XacmlCoreBase):
    """XACML PolicyDefaults type
    
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    @cvar XPATH_VERSION_ELEMENT_NAME: XML local name for XPath version element
    @type XPATH_VERSION_ELEMENT_NAME: string
   
    @ivar __xpathVersion: XPath version
    @type __xpathVersion: basestring / NoneType
    """
    ELEMENT_LOCAL_NAME = 'PolicyDefaults'
    XPATH_VERSION_ELEMENT_NAME = 'XPathVersion'
    
    __slots__ = ('__xpathVersion', )
    
    def __init__(self):
        """Initialise attributes"""
        super(PolicyDefaults, self).__init__()
        self.__xpathVersion = None
        
    def _get_xpathVersion(self):
        """@return: XPath version
        @rtype: basestring / NoneType
        """
        return self.__xpathVersion

    def _set_xpathVersion(self, value):
        """@param value: XPath version
        @type value: basestring / NoneType
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "xpathVersion" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__xpathVersion = value   

    xpathVersion = property(_get_xpathVersion, _set_xpathVersion, None, 
                            "PolicyDefaults type XPath version") 

          



"""NDG XACML Policy Information Point interface definition

"""
__author__ = "P J Kershaw"
__date__ = "15/07/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id:$"
from abc import ABCMeta, abstractmethod


class PIPInterface(object):
    """Interface class for XACML Policy Information Point.  The PDP can relay 
    attribute queries back via the Context handler to the PIP in order to help
    it in making an access control decision
    """
    __metaclass__ = ABCMeta
    __slots__ = ()
    
    @abstractmethod
    def attributeQuery(self, context, attributeDesignator):
        """Query this PIP for attributes"""
        return []


"""Utilities package for NDG XACML

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "02/04/09"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7324 2010-08-13 14:20:05Z pjkersha $'
import UserDict

# Interpret a string as a boolean
str2Bool = lambda str: str.lower() in ("yes", "true", "t", "1")

class UniqList(list):
    """Extended version of list type to enable a list with unique items.
    If an item is added that is already present then it is silently omitted
    from the list
    """
    
    def extend(self, iter):
        """Extend a list with input iterable
        @param iter: iterable object
        @type iter: iterable type
        """
        return super(UniqList, self).extend([i for i in iter if i not in self])
        
    def __iadd__(self, iter):
        """Extend a list with input iterable
        @param iter: iterable object
        @type iter: iterable type
        """
        return super(UniqList, self).__iadd__([i for i in iter 
                                               if i not in self])
         
    def append(self, item):
        """Add an item to the list
        @param item: item to append to list
        @type item: any
        """
        for i in self:
            if i == item:
                return None
            
        return super(UniqList, self).append(item)


class TypedList(list):
    """Extend list type to enabled only items of a given type.  Supports
    any type where the array type in the Standard Library is restricted to 
    only limited set of primitive types
    """
    
    def __init__(self, elementType, *arg, **kw):
        """
        @type elementType: type/tuple
        @param elementType: object type or types which the list is allowed to
        contain.  If more than one type, pass as a tuple
        """
        self.__elementType = elementType
        super(TypedList, self).__init__(*arg, **kw)

    def __repr__(self):
        """@return: represent instance giving type information about the 
        elements
        @rtype: string
        """
        return "%r type: %s" % (self.__elementType, 
                                super(TypedList, self).__repr__())
        
    def _getElementType(self):
        """@return: the element type for this list
        @rtype: type
        """
        return self.__elementType
    
    elementType = property(fget=_getElementType, 
                           doc="The allowed type or types for list elements")
     
    def extend(self, iter):
        """Extend a list with input iterable
        @param iter: iterable object
        @type iter: iterable type
        @raise TypeError: input item doesn't match list type
        """
        for i in iter:
            if not isinstance(i, self.__elementType):
                raise TypeError("List items must be of type %s" % 
                                (self.__elementType,))
                
        return super(TypedList, self).extend(iter)
        
    def __iadd__(self, iter):
        """Extend a list with input iterable
        @param iter: iterable object
        @type iter: iterable type
        @raise TypeError: input item doesn't match list type
        """
        for i in iter:
            if not isinstance(i, self.__elementType):
                raise TypeError("List items must be of type %s" % 
                                (self.__elementType,))
                    
        return super(TypedList, self).__iadd__(iter)
         
    def append(self, item):
        """Add an item to the list
        @param iter: iterable object
        @type iter: iterable type
        @raise TypeError: input item doesn't match list type
        """
        if not isinstance(item, self.__elementType):
                raise TypeError("List items must be of type %s" % 
                                (self.__elementType,))
    
        return super(TypedList, self).append(item)


class RestrictedKeyNamesDict(dict):
    """Utility class for holding a constrained list of key names
    """
    
    def __init__(self, *arg, **kw):
        """Alter standard dict() initialisation to fix key names set at 
        initialisation
        """
        super(RestrictedKeyNamesDict, self).__init__(*arg, **kw)
        self.__keyNames = self.keys() 
          
    def __setitem__(self, key, val):
        """@param key: key for item to set
        @type key: any
        @param val: value to set for this key
        @type val: any
        """
        if key not in self.__keyNames:
            raise KeyError('Key name %r not recognised.  Valid key names '
                           'are: %r' % (key, self.__keyNames))
            
        dict.__setitem__(self, key, val)

    def update(self, d, **kw):        
        """@param d: dictionary to update from
        @type d: dict
        @param kw: keywords to update dictionary with
        @type kw: dict
        """
        for dictArg in (d, kw):
            for k in dictArg:
                if k not in self.__keyNames:
                    raise KeyError('Key name "%s" not recognised.  Valid '
                                   'key names are: %s' % 
                                   self.__keyNames)
        
        dict.update(self, d, **kw)
        

_isIterable = lambda obj: getattr(obj, '__iter__', False)

  
class VettedDict(UserDict.DictMixin):
    """Enforce custom checking on keys and items before addition to a 
    dictionary
    """
    
    def __init__(self, *args):
        """Initialise setting the allowed type or types for keys and items
        
        @param args: two arguments: the first is a callable which filters for 
        permissable keys in this dict, the second sets the type or list of
        types permissable for items in this dict
        @type args: tuple
        """
        if len(args) != 2:
            raise TypeError('__init__() takes 2 arguments, KeyFilter and '
                            'valueFilter (%d given)' % len(args))
        
        # Validation of inputs
        for arg, argName in zip(args, ('KeyFilter', 'valueFilter')):
            if not callable(arg):
                raise TypeError('Expecting callable for %r input; got %r' % 
                                (argName, type(arg)))

        self.__KeyFilter, self.__valueFilter = args
        
        self.__map = {}
        
    def _verifyKeyValPair(self, key, val):
        """Check given key value pair and return False if they should be 
        filtered out.  Filter functions may also raise an exception if they
        wish to abort completely
        
        @param key: dict key to check
        @type key: basestring
        @param val: value to check
        @type val: any
        """
        if not self.__KeyFilter(key):
            return False
        
        elif not self.__valueFilter(val):
            return False
        
        else:
            return True
                  
    def __setitem__(self, key, val):
        """Enforce type checking when setting new items
        
        @param key: key for item to set
        @type key: any
        @param val: value to set for this key
        @type val: any
        """       
        if self._verifyKeyValPair(key, val):
            self.__map[key] = val

    def __getitem__(self, key):
        """Provide implementation for getting items
        @param key: key for item to retrieve
        @type key: any
        @return: value for input key
        @rtype: any
        """
        if key not in self.__map:
            raise KeyError('%r key not found in dict' % key)
        
        return self.__map[key]
    
    def get(self, key, *arg):
        """Provide implementation of get item with default
        
        @param key: key for item to retrieve
        @type key: any
        @param arg: use to set a default argument
        @type arg: tuple
        """
        if key in self.__map:
            return self.__map[key]
        
        elif len(arg) > 1:
            # Default value set
            return arg[1]
        else:
            return None
          
    def __repr__(self):
        return repr(self.__map)
    
    def keys(self):
        return self.__map.keys()
    
    def items(self):
        return self.__map.items()
    
    def values(self):
        return self.__map.values()
    
    def __contains__(self, val):
        return self.__map.__contains__(val)
    


"""NDG XACML1.0 Regular expression matching function module

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "26/03/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: regexp_match.py 7351 2010-08-20 14:46:35Z pjkersha $'
import re

from ndg.xacml.core.context.exceptions import XacmlContextTypeError
from ndg.xacml.core.attributevalue import AttributeValueClassFactory
from ndg.xacml.core.functions import (AbstractFunction, 
                                      FunctionClassFactoryInterface)


class RegexpMatchBase(AbstractFunction):
    """XACML 2.0 Regular Expression matching base class function
    
    @cvar TYPE: attribute type for the given implementation.  Derived classes
    should set appropriately
    @type TYPE: NoneType
    
    @cvar FUNCTION_SUFFIX: suffix for all regular expression type function class
    names
    @type FUNCTION_SUFFIX: string
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for regular expression type 
    function URNs
    @type FUNCTION_NS_SUFFIX: string
    
    @cvar CLASS_NAME_SUFFIX: suffix for all regular expression class names
    @type CLASS_NAME_SUFFIX: string
    """
    FUNCTION_NS = None
    FUNCTION_NS_SUFFIX = '-regexp-match'
    CLASS_NAME_SUFFIX = 'RegexpMatch'
    TYPE = None
    
    def evaluate(self, pat, input):
        """Match URI against regular expression pattern
        
        @param pat: regular expression
        @type pat: basestring
        @param input: URI to match
        @type input: type
        @return: True if URI matches pattern, False otherwise
        @rtype: bool
        """
        if not isinstance(pat, self.__class__.TYPE):
            raise TypeError('Expecting %r derived type for "pat"; got %r' %
                            (self.__class__.TYPE, type(pat)))
            
        if not isinstance(input, self.__class__.TYPE):
            raise TypeError('Expecting %r derived type for "input"; got %r' %
                            (self.__class__.TYPE, type(input)))
            
        return bool(re.match(pat.value, input.value))
    

attributeValueClassFactory = AttributeValueClassFactory()


class StringRegexMatch(RegexpMatchBase):
    """String regular expression match function class representation
    
    @cvar FUNCTION_NS: String regular expression match function URN
    @type FUNCTION_NS: string
    
    @cvar TYPE: string attribute value type
    @type TYPE: dynamically generated string type derived from 
    ndg.xacml.core.attributevalue.AttributeValue
    """
    FUNCTION_NS = 'urn:oasis:names:tc:xacml:1.0:function:string-regexp-match'
    TYPE = attributeValueClassFactory('http://www.w3.org/2001/XMLSchema#string')
    
    
class FunctionClassFactory(FunctionClassFactoryInterface):
    """Class Factory for string-regexp-match XACML function class.  Go to
    ndg.xacml.core.functions.v2.regexp_match module for additional functions 
    for other types.
    """
    def __call__(self, identifier):
        '''Create class for the Round XACML function identifier
        
        @param identifier: XACML string regular expression function identifier
        @type identifier: basestring
        @return: at least one member of class corresponding to the given input
        identifier, if no match is found returns None
        @rtype: ndg.xacml.core.functions.v1.regexp_match.StringRegexMatch / None
        '''
        if identifier == StringRegexMatch.FUNCTION_NS:
            return StringRegexMatch
        else:
            return None



"""NDG XACML Policy Decision Point type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: pdp.py 7256 2010-07-29 20:29:29Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core.context.pdpinterface import PDPInterface
from ndg.xacml.core.policy import Policy


class PDP(PDPInterface):
    """A XACML Policy Decision Point implementation.  It supports the use of a 
    single policy but not policy sets
    
    @ivar __policy: policy object for PDP to use to apply access control
    decisions
    @type policy: ndg.xacml.core.policy.Policy / None
    """
    __slots__ = ('__policy',)
    
    def __init__(self, policy=None):
        """
        @param policy: policy object for PDP to use to apply access control
        decisions, may be omitted.
        @type policy: ndg.xacml.core.policy.Policy / None
        """
        self.__policy = None
        if policy is not None:
            self.policy = policy
        
    @classmethod
    def fromPolicySource(cls, source, readerFactory):
        """Create a new PDP instance with a given policy
        @param source: source for policy
        @type source: type (dependent on the reader set, it could be for example
        a file path string, file object, XML element instance)
        @param readerFactory: reader factory returns the reader to use to read 
        this policy
        @type readerFactory: ndg.xacml.parsers.AbstractReader derived type
        """           
        pdp = cls()
        pdp.setPolicyFromSource(source, readerFactory)
        return pdp
    
    def setPolicyFromSource(self, source, readerFactory):
        """initialise PDP with the given policy
        @param source: source for policy
        @type source: type (dependent on the reader set, it could be for example
        a file path string, file object, XML element instance)
        @param readerFactory: reader factory returns the reader to use to read 
        this policy
        @type readerFactory: ndg.xacml.parsers.AbstractReader derived type
        """           
        self.policy = Policy.fromSource(source, readerFactory)
        
    @property
    def policy(self):
        """Get policy
        @return: policy object for PDP to use to apply access control decisions
        @rtype: ndg.xacml.core.policy.Policy
        """
        return self.__policy
    
    @policy.setter
    def policy(self, value):
        '''Set policy
        @param value: policy object for PDP to use to apply access control 
        decisions
        @type value: ndg.xacml.core.policy.Policy
        '''
        if not isinstance(value, Policy):
            raise TypeError('Expecting %r derived type for "policy" input; got '
                            '%r instead' % (Policy, type(value)))
        self.__policy = value
                                        
    def evaluate(self, request):
        """Make an access control decision for the given request based on the
        single policy provided
        
        @param request: XACML request context
        @type request: ndg.xacml.core.context.request.Request
        @return: XACML response instance
        @rtype: ndg.xacml.core.context.response.Response
        """
        response = self.policy.evaluate(request)
        
        return response






"""NDG XACML ElementTree Policy Reader  

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: policyreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.core.policy import Policy
from ndg.xacml.core.policydefaults import PolicyDefaults
from ndg.xacml.core.variabledefinition import VariableDefinition
from ndg.xacml.core.rule import Rule
from ndg.xacml.core.target import Target
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader
from ndg.xacml.parsers.etree.factory import ReaderFactory
    
    
class PolicyReader(ETreeAbstractReader):
    """Parse a Policy Document using ElementTree
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: type"""
    TYPE = Policy
    
    def __call__(self, obj):
        """Parse policy object
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML expression instance
        @rtype: ndg.xacml.core.policy.Policy derived type 
        @raise XMLParseError: error reading element  
        @raise NotImplementedError: parsing is not implemented for rule
        combiner, combiner parameters and obligations elements.         
        """
        elem = super(PolicyReader, self)._parse(obj)
        
        # XACML type to instantiate
        xacmlType = PolicyReader.TYPE
        policy = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
        
        # Unpack *required* attributes from top-level element
        attributeValues = []
        for attributeName in (xacmlType.POLICY_ID_ATTRIB_NAME,
                              xacmlType.RULE_COMBINING_ALG_ID_ATTRIB_NAME):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" '
                                        'element' %
                                        (attributeName,
                                         xacmlType.ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue) 
                   
        policy.policyId, policy.ruleCombiningAlgId = attributeValues
        
        # Defaults to XACML version 1.0
        # TODO: version check
        policy.version = (elem.attrib.get(xacmlType.VERSION_ATTRIB_NAME) or 
                          xacmlType.DEFAULT_XACML_VERSION)
            
        # Parse sub-elements
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            
            if localName == xacmlType.DESCRIPTION_LOCAL_NAME:
                if childElem.text is not None:
                    policy.description = childElem.text.strip()
                    
            elif localName == xacmlType.POLICY_DEFAULTS_LOCAL_NAME:
                PolicyDefaultsReader = ReaderFactory.getReader(PolicyDefaults)
                policy.policyDefaults = PolicyDefaultsReader.parse(childElem)
                   
            elif localName == Target.ELEMENT_LOCAL_NAME:
                TargetReader = ReaderFactory.getReader(Target)
                policy.target = TargetReader.parse(childElem)
             
            elif localName == xacmlType.COMBINER_PARAMETERS_LOCAL_NAME:
                raise NotImplementedError()
             
            elif localName == xacmlType.RULE_COMBINER_PARAMETERS_LOCAL_NAME:
                raise NotImplementedError()
            
            elif localName == VariableDefinition.ELEMENT_LOCAL_NAME:
                VariableDefinitionReader = ReaderFactory.getReader(
                                                            VariableDefinition)
                variableDefinition = VariableDefinitionReader.parse(childElem)
                
            elif localName == Rule.ELEMENT_LOCAL_NAME:
                RuleReader = ReaderFactory.getReader(Rule)
                policy.rules.append(RuleReader.parse(childElem))
                   
            elif localName == xacmlType.OBLIGATIONS_LOCAL_NAME:
                raise NotImplementedError('Parsing for Obligations element is '
                                          'not implemented')
            
            else:
                raise XMLParseError("XACML Policy child element name %r not "
                                    "recognised" % localName)
        
        return policy
    


"""NDG XACML one and only functions module

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "01/04/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: one_and_only.py 7087 2010-06-25 11:23:09Z pjkersha $'
#from datetime import datetime, timedelta

from ndg.xacml.core.functions import AbstractFunction, FunctionClassFactoryBase
from ndg.xacml.utils import TypedList as Bag
from ndg.xacml.core.context.exceptions import XacmlContextError


class OneAndOnlyBase(AbstractFunction):
    """Base class for XACML <type>-one-and-only functions
    
    @cvar TYPE: attribute type for the given implementation.  Derived classes
    should set appropriately
    @type TYPE: NoneType
    """
    TYPE = None
    
    def evaluate(self, bag):
        """Check a bag has one element only and return it
        
        @param bag: bag containing one element
        @type bag: ndg.xacml.utils.TypedList
        @return: single item from bag
        @rtype: dependent on bag type set in derived type
        """
        if not isinstance(bag, Bag):
            raise XacmlContextError('Expecting %r derived type for "bag"; '
                                    'got %r' % (Bag, type(bag)))
            
        if bag.elementType != self.__class__.TYPE:
            raise XacmlContextError('Expecting %r type elements for "bag"; '
                                    'got %r' %
                                    (self.__class__.TYPE, bag.elementType))
                        
        nBagElems = len(bag)
        if nBagElems != 1:
            raise XacmlContextError('Expecting single element for %r bag; got '
                                    '%r' % (self.__class__.TYPE, nBagElems))
            
        return bag[0]
    

class FunctionClassFactory(FunctionClassFactoryBase):
    """Class Factory for *-one-and-only XACML function classes
    
    @cvar FUNCTION_NAMES: one and only function URNs
    @type FUNCTION_NAMES: tuple
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for one and only function URNs
    @type FUNCTION_NS_SUFFIX: string
    
    @cvar FUNCTION_BASE_CLASS: base class for all one and only function classes 
    @type FUNCTION_BASE_CLASS: ndg.xacml.core.functions.v1.OneAndOnlyBase
    """
    FUNCTION_NAMES = (
        'urn:oasis:names:tc:xacml:1.0:function:string-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:boolean-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:integer-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:double-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:time-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:date-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:dateTime-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:anyURI-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:hexBinary-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:base64Binary-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:x500Name-one-and-only',
        'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-one-and-only'
    )
    FUNCTION_NS_SUFFIX = '-one-and-only'
    FUNCTION_BASE_CLASS = OneAndOnlyBase
    


"""NDG XACML parsers package 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "15/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7087 2010-06-25 11:23:09Z pjkersha $"
import logging
log = logging.getLogger(__name__)
from abc import ABCMeta, abstractmethod

from ndg.xacml import XacmlError
from ndg.xacml.core import XacmlCoreBase

    
class XMLParseError(XacmlError):
    """XACML package XML Parsing error"""


class AbstractReader(object):
    """Abstract base class for XACML reader"""
    __metaclass__ = ABCMeta
    
    @classmethod
    def __subclasshook__(cls, C):
        """Derived class must implement __call__"""
        if cls is AbstractReader:
            if any("__call__" in B.__dict__ for B in C.__mro__):
                return True
            
        return NotImplemented
        
    @abstractmethod
    def __call__(self, obj):
        """Abstract Parse XACML method
        @raise NotImplementedError: 
        """
        raise NotImplementedError()
    
    @classmethod
    def parse(cls, obj):
        """Parse from input object and return new XACML object
        @param obj: input source - file name, stream object or other
        @type obj: string, stream or other
        @return: new XACML object
        @rtype: XacmlCoreBase sub type
        """
        reader = cls()
        return reader(obj)
    
    
class AbstractReaderFactory(object):
    """Abstract base class XACML reader factory"""
    __metaclass__ = ABCMeta
    
    @classmethod
    @abstractmethod
    def getReader(cls, xacmlType):
        """Get the reader class for the given XACML input type
        @param xacmlType: XACML type to retrieve a reader for
        @type xacmlType: ndg.xaml.core.XacmlCoreBase derived
        @return: reader class
        @rtype: ndg.xacml.parsers.AbstractReader derived type
        """
        if not issubclass(xacmlType, XacmlCoreBase):
            raise TypeError('Expecting %r derived class for getReader method; '
                            'got %r' % (XacmlCoreBase, xacmlType))



"""NDG Security Condition type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "19/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: apply.py 7365 2010-08-25 15:41:20Z pjkersha $"
from ndg.xacml.utils import TypedList
from ndg.xacml.core.expression import Expression
from ndg.xacml.core.functions import (FunctionMap, UnsupportedStdFunctionError,
                                      UnsupportedFunctionError)


class Apply(Expression):
    """XACML Apply type
    
    @cvar ELEMENT_LOCAL_NAME: XML element local name
    @type ELEMENT_LOCAL_NAME: string
    @cvar FUNCTION_ID_ATTRIB_NAME: function ID XML attribute name
    @type FUNCTION_ID_ATTRIB_NAME: string

    @ivar __functionId: URN corresponding to function to be applied
    @type __functionId: basestring/NoneType
    @ivar __function: function to be applied
    @type __function: ndg.xacml.core.functions.AbstractFunction derived type
    @ivar __functionMap: function mapping object to map URNs to function class
    implementations
    @type __functionMap: ndg.xacml.core.functions.FunctionMap
    @ivar __loadFunctionFromId: boolean determines whether or not to load
    function classes for given function URN in functionId set property method
    @type __loadFunctionFromId: bool
    @ivar __expressions: list of expressions contained in the Apply statement
    @type __expressions: ndg.xacml.utils.TypedList
    """
    ELEMENT_LOCAL_NAME = 'Apply'
    FUNCTION_ID_ATTRIB_NAME = 'FunctionId'
    
    __slots__ = (
        '__functionId', 
        '__function', 
        '__functionMap',
        '__loadFunctionFromId',
        '__expressions'
    )
    
    def __init__(self):
        """Initialise attributes"""
        super(Apply, self).__init__()
        self.__functionId = None
        self.__function = None
        self.__functionMap = FunctionMap()
        self.__loadFunctionFromId = True
        self.__expressions = TypedList(Expression)
      
    @property
    def loadFunctionFromId(self):
        """Set to False to stop the functionId property set method automatically
        trying to load the corresponding function for the given functionId
        
        @return: flag setting
        @rtype: bool""" 
        return self.__loadFunctionFromId
    
    @loadFunctionFromId.setter
    def loadFunctionFromId(self, value):
        """Set to False to stop the functionId property set method automatically
        trying to load the corresponding function for the given functionId
        
        @param value: flag setting
        @type value: bool
        @raise TypeError: incorrect input type
        """ 
        if not isinstance(value, bool):
            raise TypeError('Expecting %r type for "loadFunctionFromId" '
                            'attribute; got %r' % (bool, type(value)))
            
        self.__loadFunctionFromId = value
        
    def _get_functionId(self):
        """Get function ID
        @return: function ID for this Apply statement
        @rtype: basestring/NoneType
        """
        return self.__functionId

    def _set_functionId(self, value):
        """Set function ID
        @param value: function URN
        @type value: basestring
        @raise TypeError: incorrect input type
        """ 
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "functionId" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__functionId = value
        
        # Also retrieve function for this function ID if a map has been set
        if self.__loadFunctionFromId:
            self.setFunctionFromMap(self.__functionMap)   

    functionId = property(_get_functionId, _set_functionId, None, 
                          "Apply type Function ID") 
        
    def setFunctionFromMap(self, functionMap):
        """Set the function from a function map - a dictionary of function ID to
        function mappings.  The function is looked up based on the "functionId"
        attribute.  This method is automatically called when the functionId set
        property method is invoked.  To switch off this behaviour set
        
        loadFunctionFromId = False
    
        @param functionMap: function mapping object to map URNs to function 
        class implementations
        @type functionMap: ndg.xacml.core.functions.FunctionMap
        
        @raise UnsupportedStdFunctionError: policy references a function type 
        which is in the XACML spec. but is not supported by this implementation
        @raise UnsupportedFunctionError: policy references a function type which
        is not supported by this implementation
        """
        if self.functionId is None:
            raise AttributeError('"functionId" attribute must be set in order '
                                 'to retrieve the required function')
            
        # Get function class for this <Apply> statement         
        functionClass = functionMap.get(self.functionId)
        if functionClass is NotImplemented:
            raise UnsupportedStdFunctionError('No match function class '
                                              'implemented for MatchId="%s"' % 
                                              self.functionId)
        elif functionClass is None:
            raise UnsupportedFunctionError('<Apply> function namespace %r is '
                                           'not recognised' % 
                                           self.functionId) 
            
        self.__function = functionClass()
    
    @property
    def functionMap(self):
        """functionMap object for PDP to retrieve functions from given XACML
        function URNs
        @return: function mapping object
        @rtype: ndg.xacml.core.functions.FunctionMap
        """
        return self.__functionMap
    
    @functionMap.setter
    def functionMap(self, value):
        '''functionMap object for PDP to retrieve functions from given XACML
        function URNs
        @param value: function mapping object to map URNs to function 
        class implementations
        @type value: ndg.xacml.core.functions.FunctionMap
        @raise TypeError: incorrect type for input value
        '''
        if not isinstance(value, FunctionMap):
            raise TypeError('Expecting %r derived type for "functionMap" '
                            'input; got %r instead' % (FunctionMap, 
                                                       type(value)))
        self.__functionMap = value
          
    @property  
    def function(self):
        """Get Function for this <Apply> instance
        @return: function to be applied
        @rtype: ndg.xacml.core.functions.AbstractFunction derived type
        """
        return self.__function
        
    @property
    def expressions(self):
        """List of expression sub-elements
        @return: list of expressions contained in the Apply statement
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__expressions 
    
    def evaluate(self, context):
        """Evaluate a given <Apply> statement in a rule condition
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: attribute value(s) resulting from execution of this expression
        in a condition
        @rtype: AttributeValue/NoneType
        """ 
        
        # Marshal inputs
        funcInputs = [None]*len(self.expressions)

        for i, expression in enumerate(self.expressions):
            funcInputs[i] = expression.evaluate(context)
            
        # Execute function on the retrieved inputs
        result = self.function.evaluate(*tuple(funcInputs))
        
        # Pass the result back to the parent <Apply> element
        return result
          


"""NDG XACML ElementTree based reader for subject match type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: subjectmatchreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.match import SubjectMatch
from ndg.xacml.core.attributedesignator import SubjectAttributeDesignator
from ndg.xacml.parsers.etree.matchreader import MatchReaderBase


class SubjectMatchReader(MatchReaderBase):
    """ElementTree based parser for XACML SubjectMatch
    
    @cvar TYPE: XACML class type that this reader will read values into
    @type TYPE: abc.ABCMeta
    
    @cvar ATTRIBUTE_DESIGNATOR_TYPE: type for attribute designator sub-elements
    @type ATTRIBUTE_DESIGNATOR_TYPE: abc.ABCMeta
    """
    TYPE = SubjectMatch
    ATTRIBUTE_DESIGNATOR_TYPE = SubjectAttributeDesignator


"""NDG XACML ElementTree reader module containing reader base class 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "19/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: factory.py 7087 2010-06-25 11:23:09Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.parsers import AbstractReaderFactory
from ndg.xacml.utils.factory import importModuleObject

from ndg.xacml.core.policy import Policy
from ndg.xacml.core.target import Target
from ndg.xacml.core.rule import Rule

# Target child elements
from ndg.xacml.core.subject import Subject
from ndg.xacml.core.action import Action
from ndg.xacml.core.resource import Resource
from ndg.xacml.core.environment import Environment


class ReaderFactory(AbstractReaderFactory):
    """Parser factory for ElementTree based parsers for XACML types"""
    
    @classmethod
    def getReader(cls, xacmlType):
        """Return ElementTree based Reader class for the given input
        
        @param xacmlType: XACML type to return a parser class for
        @type xacmlType: type
        @return: ElementTree based reader for the input XACML type.  The class
        and module containing the class are infered from the XACML class name
        input e.g. 
        
        ndg.xacml.core.Subject => ndg.xacml.parsers.etree.subjectreader.SubjectReader
        
        @rtype: ndg.xacml.parsers.etree.reader.ETreeAbstractReader derived
        type
        @raise ImportError: if no reader class found for input type
        """
        xacmlTypeName = xacmlType.__name__
        readerClassName = 'ndg.xacml.parsers.etree.%sreader.%sReader' % (
                                                        xacmlTypeName.lower(),
                                                        xacmlTypeName)
        readerClass = importModuleObject(readerClassName)
        return readerClass
            



"""NDG XACML string bag function module

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "31/03/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: bag.py 7109 2010-06-28 12:54:57Z pjkersha $'
from ndg.xacml.utils import TypedList as Bag
from ndg.xacml.core.functions import AbstractFunction, FunctionClassFactoryBase


class BagBase(AbstractFunction):
    """Abstract bag function for all types of bag function
    
    urn:oasis:names:tc:xacml:1.0:function:<type>-bag
    
    @cvar TYPE: attribute type for the given implementation.  Derived classes
    should set appropriately
    @type TYPE: NoneType
    
    @cvar FUNCTION_SUFFIX: suffix for all bag function class names
    @type FUNCTION_SUFFIX: string
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for at least one member of function
    URNs
    @type FUNCTION_NS_SUFFIX: string
    """
    TYPE = None
    FUNCTION_SUFFIX = 'Bag'
    FUNCTION_NS_SUFFIX = '-bag'
    
    def evaluate(self, *args):
        """return inputs into a bag
        
        @param args: inputs to add to bag
        @type args: tuple
        @return: True if strings match, False otherwise
        @rtype: bool
        """
        bag = Bag(self.__class__.TYPE)
        for i in args:
            if not isinstance(i, self.__class__.TYPE):
                raise TypeError('Expecting %r derived type for bag element; '
                                'got %r' % (self.__class__.TYPE, type(i)))
            bag.append(i)
         
        return bag


class FunctionClassFactory(FunctionClassFactoryBase):
    """Class Factory for *-bag XACML function classes
    
    @cvar FUNCTION_NAMES: bag function URNs
    @type FUNCTION_NAMES: tuple
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for bag function URNs
    @type FUNCTION_NS_SUFFIX: string
    
    @cvar FUNCTION_BASE_CLASS: base class for all bag function classes
    @type FUNCTION_BASE_CLASS: ndg.xacml.core.functions.v1.BagBase
    """
    FUNCTION_NAMES = (
        'urn:oasis:names:tc:xacml:1.0:function:string-bag',
        'urn:oasis:names:tc:xacml:1.0:function:boolean-bag',
        'urn:oasis:names:tc:xacml:1.0:function:integer-bag',
        'urn:oasis:names:tc:xacml:1.0:function:double-bag',
        'urn:oasis:names:tc:xacml:1.0:function:time-bag',
        'urn:oasis:names:tc:xacml:1.0:function:date-bag',
        'urn:oasis:names:tc:xacml:1.0:function:dateTime-bag',
        'urn:oasis:names:tc:xacml:1.0:function:anyURI-bag',
        'urn:oasis:names:tc:xacml:1.0:function:hexBinary-bag',
        'urn:oasis:names:tc:xacml:1.0:function:base64Binary-bag',
        'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-bag',
        'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-bag',
        'urn:oasis:names:tc:xacml:1.0:function:x500Name-bag',
        'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-bag',
    )
    FUNCTION_NS_SUFFIX = '-bag'
    FUNCTION_BASE_CLASS = BagBase 



"""NDG Security Variable Reference type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "29/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: variablereference.py 7108 2010-06-28 09:19:26Z pjkersha $"
from ndg.xacml.core import XacmlPolicyBase


class VariableReference(XacmlPolicyBase):
    '''XACML Variable Reference Type - this class is a placeholder, it's not 
    currently implemented
    
    @cvar ELEMENT_LOCAL_NAME: XML element local name
    @type ELEMENT_LOCAL_NAME: string
    '''
    ELEMENT_LOCAL_NAME = "VariableReference"

    __slots__ = ()
    
    def __init__(self):
        '''This class not needed yet'''
        raise NotImplementedError()


"""NDG XACML URI equal matching function module

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "30/03/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: at_least_one_member_of.py 7351 2010-08-20 14:46:35Z pjkersha $'
from ndg.xacml.core.functions import AbstractFunction, FunctionClassFactoryBase
from ndg.xacml.core.context.exceptions import XacmlContextTypeError
from ndg.xacml.utils import TypedList as Bag


class AtLeastOneMemberOfBase(AbstractFunction):
    """Base class implementation for at least one member of XACML function -
    check at least one item in one set is contained in the second set
    
    urn:oasis:names:tc:xacml:1.0:function:<type>-at-least-one-member-of
    
    @cvar TYPE: attribute type for the given implementation.  Derived classes
    should set appropriately
    @type TYPE: NoneType
    """
    TYPE = None
    
    def evaluate(self, set1, set2):
        """Check input is contained in the bag
        
        @param set1: check to see if at least one item in this set is contained 
        in the second set
        @type set1: TypedList(self.__class__.TYPE)
        @param set2: bag of self.__class__.TYPE values
        @type set2: TypedList(self.__class__.TYPE)
        @return: True if str is in bag, False otherwise
        @rtype: bool
        """
        if not isinstance(set1, Bag):
            raise XacmlContextTypeError('Expecting %r derived type for "set1"; '
                                        'got %r' % (Bag, type(set1)))
            
        if set1.elementType != self.__class__.TYPE:
            raise XacmlContextTypeError('Expecting %r type elements for '
                                        '"set1"; got %r' %
                                        (self.__class__.TYPE, set1.elementType))
            
        if not isinstance(set2, Bag):
            raise XacmlContextTypeError('Expecting %r derived type for "set2"; '
                                        'got %r' % (Bag, type(set2)))
            
        if set2.elementType != self.__class__.TYPE:
            raise XacmlContextTypeError('Expecting %r type elements for '
                                        '"set2"; got %r' %
                                        (self.__class__.TYPE, set2.elementType))
            
        _set1 = set([attr.value for attr in set1])
        _set2 = set([attr.value for attr in set2])
        
        return len(list(_set1 & _set2)) > 0
    

class FunctionClassFactory(FunctionClassFactoryBase):
    """Class Factory for *-at-least-one-member-of XACML function classes
    
    @cvar FUNCTION_NAMES: list of at least one member of function URNs
    @type FUNCTION_NAMES: tuple
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for at least one member of function
    URNs
    @type FUNCTION_NS_SUFFIX: string
    
    @cvar FUNCTION_BASE_CLASS: base class for all at least one member of 
    function classes
    @type FUNCTION_BASE_CLASS: ndg.xacml.core.functions.v1.AtLeastOneMemberOfBase
    """
    FUNCTION_NAMES = (
'urn:oasis:names:tc:xacml:1.0:function:string-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:boolean-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:integer-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:double-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:time-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:date-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-at-least-one-member-of',
    )
    FUNCTION_NS_SUFFIX = '-at-least-one-member-of'
    FUNCTION_BASE_CLASS = AtLeastOneMemberOfBase



"""NDG XACML ElementTree based reader for AttributeValue type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: attributevaluereader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.attributevalue import (AttributeValue, 
                                           AttributeValueClassFactory)
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.expressionreader import ExpressionReader


class AttributeValueReader(ExpressionReader):
    '''ElementTree based XACML Expression type parser
    
    @cvar TYPE: XACML class type that this reader will read values into
    @type TYPE: abc.ABCMeta
    
    @cvar FACTORY: factory function for returning an Attribute value type for a 
    given XACML Attribute value URI
    @type FACTORY: ndg.xacml.core.attributevalue.AttributeValueClassFactory
    '''
    TYPE = AttributeValue
    FACTORY = AttributeValueClassFactory()
    
    def __call__(self, obj):
        """Parse *AttributeValue type element - override this method instead of
        _parseExtension since AttributeValue class is virtual.  A sub-type can
        be instantiated only once the data type attribute is parsed
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML attribute value instance
        @rtype: ndg.xacml.core.attributevalue.AttributeValue derived type 
        @raise XMLParseError: error reading element       
        """
        elem = super(AttributeValueReader, self)._parse(obj)
        
        xacmlType = self.__class__.TYPE
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
            
        # Unpack *required* attributes from top-level element
        elemAttributeValues = []
        for attributeName in (xacmlType.DATA_TYPE_ATTRIB_NAME,):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" element' %
                                    (attributeName, 
                                     xacmlType.ELEMENT_LOCAL_NAME))
                
            elemAttributeValues.append(attributeValue)
             
        attributeValueClass = self.__class__.FACTORY(elemAttributeValues[0])
        attributeValue = attributeValueClass()
        attributeValue.dataType = elemAttributeValues[0]
        self._parseExtension(elem, attributeValue)
        
        return attributeValue
    
    def _parseExtension(self, elem, attributeValue):
        """Parse XML Attribute value element
        
        @param elem: ElementTree XML element
        @type elem: xml.etree.Element
        
        @param attributeValue: attribute selector to be updated with parsed
        values
        @type attributeValue: ndg.xacml.core.attributevalue.AttributeValue
        
        @raise XMLParseError: error parsing attribute ID XML attribute
        """
        if elem.text is None:
            raise XMLParseError('No attribute value element found parsing %r' % 
                                AttributeValueReader.TYPE.ELEMENT_LOCAL_NAME) 
            
        attributeValue.value = elem.text
    


"""NDG Security Rule type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: rule.py 7365 2010-08-25 15:41:20Z pjkersha $"
import traceback
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.target import Target
from ndg.xacml.core.condition import Condition
from ndg.xacml.core.context.result import Decision


class Effect(object):
    """Rule Effect
    
    @cvar PERMIT_STR: permit decision string
    @type PERMIT_STR: string
    
    @cvar DENY_STR: deny decision string
    @type DENY_STR: string
    
    @cvar TYPES: list of valid effect strings
    @type TYPES: tuple
    
    @ivar value: effect value
    @type value: string
    """
    DENY_STR = 'Deny'
    PERMIT_STR = 'Permit'
    TYPES = (DENY_STR, PERMIT_STR)
    __slots__ = ('__value',)
    
    def __init__(self, effect=DENY_STR):
        """@param effect: initialise effect value, defaults to deny
        @type effect: basestring / ndg.xacml.core.rule.Effect
        """
        self.__value = None
        self.value = effect

    def __getstate__(self):
        '''Enable pickling
        
        @return: class instance attributes dictionary
        @rtype: dict
        '''
        _dict = {}
        for attrName in Effect.__slots__:
            # Ugly hack to allow for derived classes setting private member
            # variables
            if attrName.startswith('__'):
                attrName = "_Effect" + attrName
                
            _dict[attrName] = getattr(self, attrName)
            
        return _dict
  
    def __setstate__(self, attrDict):
        '''Enable pickling
        
        @param attrDict: class instance attributes dictionary
        @type attrDict: dict
        '''
        for attrName, val in attrDict.items():
            setattr(self, attrName, val)
            
    def _setValue(self, value):
        """Set effect value
        
        @param value: effect value - constrained vocabulary to Effect.TYPES
        @type value: string or ndg.xacml.core.rule.Effect
        @raise AttributeError: invalid decision string value input
        @raise TypeError: invalid type for input decision value
        """
        if isinstance(value, Effect):
            # Cast to string
            value = str(value)
            
        elif not isinstance(value, basestring):
            raise TypeError('Expecting string or Effect instance for '
                            '"value" attribute; got %r instead' % type(value))
            
        if value not in self.__class__.TYPES:
            raise AttributeError('Permissable effect types are %r; got '
                                 '%r instead' % (Effect.TYPES, value))
        self.__value = value
        
    def _getValue(self):
        """Get effect value
        
        @return: effect value 
        @rtype: string
        """
        return self.__value
    
    value = property(fget=_getValue, fset=_setValue, doc="Effect value")
    
    def __str__(self):
        """represent decision as a string
        
        @return: decision value 
        @rtype: string 
        """
        return self.__value

    def __eq__(self, effect):
        """
        @param effect: effect value to compare with self's
        @type effect: string or ndg.xacml.core.rule.Effect
        @return: True if the decision values match, False otherwise
        @rtype: bool
        @raise AttributeError: invalid decision string value input
        @raise TypeError: invalid type for input decision value
        """
        if isinstance(effect, Effect):
            # Cast to string
            value = effect.value
            
        elif isinstance(effect, basestring):
            value = effect
            
        else:
            raise TypeError('Expecting string or Effect instance for '
                            'input effect value; got %r instead' % type(value))
            
        if value not in self.__class__.TYPES:
            raise AttributeError('Permissable effect types are %r; got '
                                 '%r instead' % (Effect.TYPES, value))
            
        return self.__value == value 
    
    def __nonzero__(self):
        """Boolean evaluation of a rule effect - True = Allow; False = Deny
        
        @return: True if the effect value is permit, False otherwise
        @rtype: bool
        """
        return self.__value == Effect.PERMIT_STR       


class PermitEffect(Effect):
    """Permit authorisation Effect"""
    __slots__ = ()

    def __init__(self):
        """Initialise set with Permit value"""
        super(PermitEffect, self).__init__(Effect.PERMIT_STR)
        
    def _setValue(self, value):  
        """Make value read-only
        @raise AttributeError: value can't be set
        """
        raise AttributeError("can't set attribute")


class DenyEffect(Effect):
    """Deny authorisation Effect"""
    __slots__ = ()
    
    def __init__(self):
        """Initialise set with Permit value"""
        super(DenyEffect, self).__init__(Effect.DENY_STR)
        
    def _setValue(self, value):  
        """Make value read-only
        @raise AttributeError: value can't be set
        """
        raise AttributeError("can't set attribute")

# Add instances of each for convenience
Effect.PERMIT = PermitEffect()
Effect.DENY = DenyEffect()


class Rule(XacmlCoreBase):
    """XACML Policy Rule
    
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    @cvar DESCRIPTION_LOCAL_NAME: XML local name for the description element
    @type DESCRIPTION_LOCAL_NAME: string
    @cvar RULE_ID_ATTRIB_NAME: rule id XML attribute name
    @type RULE_ID_ATTRIB_NAME: string
    @cvar EFFECT_ATTRIB_NAME: effect XML attribute name
    @type EFFECT_ATTRIB_NAME: string

    @ivar __target: rule target
    @type __target: ndg.xacml.core.target.Target / NoneType
    @ivar __condition: rule condition
    @type __condition: ndg.xacml.core.condition.Condition / NoneType
    @ivar __description: rule description text
    @type __description: basestring / NoneType
    @ivar __id: rule ID
    @type __id: basestring / NoneType
    @ivar __effect: rule effect
    @type __effect: ndg.xacml.core.rule.Effect / NoneType
    """
    ELEMENT_LOCAL_NAME = 'Rule'
    RULE_ID_ATTRIB_NAME = 'RuleId'
    EFFECT_ATTRIB_NAME = 'Effect'
    
    DESCRIPTION_LOCAL_NAME = 'Description'
    
    __slots__ = (
        '__target', 
        '__condition', 
        '__description', 
        '__id', 
        '__effect'
    )
    
    def __init__(self):
        """Initialise attributes"""
        super(Rule, self).__init__()
        
        self.__id = None
        self.__effect = None
        self.__target = None
        self.__condition = None
        
    @property
    def target(self):
        """Get Rule target
        @return: rule target
        @rtype: ndg.xacml.core.target import Target / NoneType 
        """
        return self.__target
    
    @target.setter
    def target(self, value):
        """Set rule target
        @param value: rule target
        @type value: ndg.xacml.core.target import Target 
        @raise TypeError: incorrect type set
        """
        if not isinstance(value, Target):
            raise TypeError('Expecting %r type for "id" '
                            'attribute; got %r' % (Target, type(value))) 
        self.__target = value
           
    @property
    def condition(self):
        """Get rule condition
        
        @return: rule condition
        @rtype: ndg.xacml.core.condition.Condition / NoneType
        """
        return self.__condition
    
    @condition.setter
    def condition(self, value):
        """Set rule condition
        
        @param value: rule condition
        @type value: ndg.xacml.core.condition.Condition
        @raise TypeError: incorrect type set
        """
        if not isinstance(value, Condition):
            raise TypeError('Expecting %r type for "id" attribute; got %r' % 
                            (Condition, type(value))) 
            
        self.__condition = value
             
    def _get_id(self):
        """Get rule ID
        
        @return: rule ID
        @rtype: ndg.xacml.core.condition.Condition / NoneType
        """
        return self.__id

    def _set_id(self, value):
        """Set rule ID
        
        @param value: rule ID
        @type value: basestring
        @raise TypeError: incorrect type set
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "id" attribute; got %r' % 
                            (basestring, type(value)))
         
        self.__id = value   

    id = property(_get_id, _set_id, None, "Rule identifier attribute")  
      
    def _get_effect(self):
        """Get rule effect
        
        @return: rule effect
        @rtype: ndg.xacml.core.rule.Effect / NoneType
        """
        return self.__effect

    def _set_effect(self, value):
        """Set rule effect
        
        @param value: rule effect
        @type value: ndg.xacml.core.rule.Effect
        @raise TypeError: incorrect type set
        """
        if not isinstance(value, Effect):
            raise TypeError('Expecting %r type for "effect" '
                            'attribute; got %r' % (Effect, type(value)))
            
        self.__effect = value   

    effect = property(_get_effect, _set_effect, None, 
                      "Rule effect attribute")  

    def _getDescription(self):
        """Get rule description
        
        @return: rule description
        @rtype: basestring / NoneType
        """
        return self.__description

    def _setDescription(self, value):
        """Set rule description
        
        @param value: rule description
        @type value: basestring
        @raise TypeError: incorrect type set
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "description" '
                            'attribute; got %r' % type(value))
        self.__description = value

    description = property(_getDescription, _setDescription, 
                           doc="Rule Description text")
    
    def evaluate(self, context):
        """Evaluate a rule
        @param context: the request context
        @type context: ndg.xacml.core.request.Request
        @return: result of the evaluation - the decision for this rule
        @rtype: ndg.xacml.core.context.result.Decision
        """
        
        # Place exception block to enable rule combining algorithm which calls
        # this method to correctly handle Indeterminate results
        try:
            log.debug('Evaluating rule %r ...', self.id)
            
            # Instantiation implicitly sets to default value of Indeterminate
            decision = Decision()
            
            # Check for a rule target
            if self.target is not None:
                targetMatch = self.target.match(context)
                if targetMatch:
                    log.debug('Match to request context for target in rule '
                              '%r', self.id)
            else:
                log.debug('No target set in rule %r', self.id)
                targetMatch = True
      
            if not targetMatch:
                log.debug('No match to request context for target in rule '
                          '%r returning NotApplicable status', self.id)
                decision = Decision.NOT_APPLICABLE
                return decision
            
            # Apply the condition if present
            if self.condition is not None:
                conditionStatus = self.condition.evaluate(context)
            else:
                # No condition set to True - 7.8 in spec.:
                #
                # The condition value SHALL be "True" if the <Condition> element
                # is absent
                log.debug('No condition set for rule %r: setting condition '
                          'status True', self.id)
                conditionStatus = True
                
            # Ref. Spec. 7.9 Rule evaluation, Nb. to get this far, the target
            # must evaluated as True
            if conditionStatus:
                decision = Decision(decision=self.effect.value)
            else:
                decision = Decision.NOT_APPLICABLE
                
            return decision
        
        except Exception, e:
            log.error('Error occurred evaluating rule %r, returning '
                      'Indeterminate result to caller: %s',
                      self.id,
                      traceback.format_exc())
            return Decision.INDETERMINATE



"""NDG XACML ElementTree based Generic Target Child Element reader - for
Reosurce, Subject, Action and Environment

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "18/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: targetchildreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader
from ndg.xacml.parsers.etree.factory import ReaderFactory


class TargetChildReader(ETreeAbstractReader):
    '''ElementTree based XACML generic target child element parser
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: type
    '''

    def __call__(self, obj):
        """Parse target child element object
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML expression instance
        @rtype: ndg.xacml.core.TargetChildBase derived type 
        @raise XMLParseError: error reading element               
        """
        elem = super(TargetChildReader, self)._parse(obj)
        
        xacmlType = self.__class__.TYPE
        targetChild = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
            
        # Parse match elements
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            
            if localName == xacmlType.MATCH_TYPE.ELEMENT_LOCAL_NAME:
                # Get reader for the match type
                matchReader = ReaderFactory.getReader(xacmlType.MATCH_TYPE)
                targetChild.matches.append(matchReader.parse(childElem))
            
            else:
                raise XMLParseError("XACML %r child element name %r not "
                                    "recognised" % (xacmlType.ELEMENT_LOCAL_NAME, 
                                                    localName))
                       
        return targetChild


"""NDG XACML ElementTree based reader for PolicyDefaults type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "18/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: policydefaultsreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.policydefaults import PolicyDefaults
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader


class PolicyDefaultsReader(ETreeAbstractReader):
    '''ElementTree based XACML PolicyDefaults type parser
    
    @cvar TYPE: XACML class type that this reader will read values into
    @type TYPE: type
    '''
    TYPE = PolicyDefaults
    
    def __call__(self, obj):
        """Parse Policy defaults object
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML expression instance
        @rtype: ndg.xacml.core.policydefaults.PolicyDefaults derived type 
        @raise XMLParseError: error reading element               
        """
        elem = super(PolicyDefaultsReader, self)._parse(obj)
        
        xacmlType = self.__class__.TYPE
        policyDefaults = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
        
        if len(elem) != 1:
            raise XMLParseError('Expecting a single child element for '
                                'PolicyDefaults element')
            
        if (QName.getLocalPart(elem[0].tag) != 
            xacmlType.XPATH_VERSION_ELEMENT_NAME):
            raise XMLParseError('Expecting a %r child element for '
                                'PolicyDefaults element' % 
                                xacmlType.XPATH_VERSION_ELEMENT_NAME)
        
        xpathVersion = elem[0].text 
        if xpathVersion is None:
            raise XMLParseError('No %r child element value set for '
                                'PolicyDefaults element' % 
                                xacmlType.XPATH_VERSION_ELEMENT_NAME) 
             
        policyDefaults.xpathVersion = xpathVersion
        
        return policyDefaults





"""NDG XACML exception types for the context package

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "01/04/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: exceptions.py 7087 2010-06-25 11:23:09Z pjkersha $"
from ndg.xacml.core.context.response import Response
from ndg.xacml.core.context.result import Result, Decision, StatusCode


class XacmlContextError(Exception):    
    """Base class for exceptions related XACML context handling
    
    @ivar __response: Context response object associated with this exception
    @type __response: ndg.xacml.core.context.response.Response
    """
    
    def __init__(self, *arg, **kw):
        """Override Exception base class so as to add the response object in
        the exception instance
        
        @param arg: base class arguments
        @type arg: tuple
        @param kw: base class keywords if any
        @type kw: dict
        """
        super(XacmlContextError, self).__init__(*arg, **kw)
        self.__response = Response()
        self.__response.results.append(Result.createInitialised())
        
        self.__response.results[0].decision = Decision.INDETERMINATE
        if len(arg) > 0:
            self.__response.results[0].status.statusMessage = arg[0]
        
    @property
    def response(self):
        """
        @return: Context response object associated with this exception
        @rtype: ndg.xacml.core.context.response.Response
        """
        return self.__response
    
    @response.setter
    def response(self, value):
        if not isinstance(value, Response):
            raise TypeError('Expecting %r type for "response" attribute; got '
                            '%r instead' % (Response, type(Response)))
        self.__response = value


class XacmlContextTypeError(XacmlContextError):
    """Type errors within XACML context processing"""
            
        
class MissingAttributeError(XacmlContextError):
    """AttributeDesignator or AttributeSelector has specified MustBePresent
    but the request context contains no match for required attribute XPath
    respectively
    """



"""NDG XACML ElementTree based Target Element reader 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: targetreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.target import Target
from ndg.xacml.core.subject import Subject
from ndg.xacml.core.resource import Resource
from ndg.xacml.core.action import Action
from ndg.xacml.core.environment import Environment
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader
from ndg.xacml.parsers.etree.factory import ReaderFactory


class TargetReader(ETreeAbstractReader):
    """ElementTree based parser for XACML Target elements
    
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: type
    """
    TYPE = Target
    
    def __call__(self, obj):
        """Parse Target object
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML expression instance
        @rtype: ndg.xacml.core.target.Target derived type 
        @raise XMLParseError: error reading element               
        """
        elem = super(TargetReader, self)._parse(obj)
        
        xacmlType = TargetReader.TYPE
        target = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError('No "%s" element found' % 
                                xacmlType.ELEMENT_LOCAL_NAME)
        
        # Parse sub-elements
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            
            if localName == xacmlType.SUBJECTS_ELEMENT_LOCAL_NAME:
                for subjElem in childElem:
                    SubjectReader = ReaderFactory.getReader(Subject)
                    target.subjects.append(SubjectReader.parse(subjElem))
                                
            elif localName == xacmlType.RESOURCES_ELEMENT_LOCAL_NAME:
                for resourceElem in childElem:
                    ResourceReader = ReaderFactory.getReader(Resource)
                    target.resources.append(ResourceReader.parse(resourceElem))
                
            elif localName == xacmlType.ACTIONS_ELEMENT_LOCAL_NAME:
                for targetElem in childElem:
                    ActionReader = ReaderFactory.getReader(Action)
                    target.actions.append(ActionReader.parse(targetElem))
                
            elif localName == xacmlType.ENVIRONMENTS_ELEMENT_LOCAL_NAME:
                for environElem in childElem:
                    EnvironmentReader = ReaderFactory.getReader(Environment)
                    target.environments.append(EnvironmentReader.parse(
                                                                environElem))
            else:
                raise XMLParseError("XACML Target child element name %r not "
                                    "recognised" % localName)
                
        return target


"""NDG Security Context handler base class

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: handler.py 7336 2010-08-18 09:07:40Z pjkersha $"
from ndg.xacml.core.context.handlerinterface import CtxHandlerInterface
from ndg.xacml.core.context.pdpinterface import PDPInterface
from ndg.xacml.core.context.pipinterface import PIPInterface
    

class CtxHandlerBase(CtxHandlerInterface):
    """Base class for Context handlers - extends Context handler interface to 
    include Policy Decision Point and Policy Information Point references
    """
    
    __slots__ = (
        '__pip',
        '__pdp', 
    )
      
    def __init__(self):
        self.__pip = None
        self.__pdp = None
        
    def _getPip(self):
        return self.__pip

    def _setPip(self, value):
        if not isinstance(value, PIPInterface):
            raise TypeError('Expecting %r type for "pip" attribute; got %r '
                            'instead' % 
                            (PIPInterface, value))
            
        self.__pip = value

    pip = property(_getPip, _setPip, None, "Policy Information Point")
          
    def _getPdp(self):
        return self.__pdp

    def _setPdp(self, value):
        if not isinstance(value, PDPInterface):
            raise TypeError('Expecting %r type for "pdp" attribute; got %r '
                            'instead' % 
                            (PDPInterface, value))
            
        self.__pdp = value

    pdp = property(_getPdp, _setPdp, None, "Policy Decision Point")
        


"""NDG XACML ElementTree Attribute Selector Reader  

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "18/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: attributeselectorreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.utils import str2Bool
from ndg.xacml.core.attributeselector import AttributeSelector
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.expressionreader import ExpressionReader


class AttributeSelectorReader(ExpressionReader):
    '''ElementTree based parser for XACML Attribute Selector type
    
    @cvar TYPE: XACML class type that this reader will read values into
    @type TYPE: abc.ABCMeta
    '''
    TYPE = AttributeSelector
    
    def _parseExtension(self, elem, attributeSelector):
        """Parse XML Attribute Selector element
        
        @param elem: ElementTree XML element
        @type elem: xml.etree.Element
        
        @param attributeSelector: attribute selector to be updated with parsed
        values
        @type attributeSelector: ndg.xacml.core.attributeSelector.AttributeSelector
        
        @raise XMLParseError: error parsing attribute ID XML attribute
        
        @return: updated attribute selector
        @rtype: ndg.xacml.core.attributeSelector.AttributeSelector
        """
        
        xacmlType = self.__class__.TYPE
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
            
        # Unpack *required* attributes from top-level element
        attributeValues = []
        for attributeName in (xacmlType.REQUEST_CONTEXT_PATH_ATTRIB_NAME,):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" element' %
                                    (attributeName, 
                                     xacmlType.ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
             
        attributeSelector.requestContextPath, = attributeValues
        
        mustBePresent = elem.attrib.get(xacmlType.MUST_BE_PRESENT_ATTRIB_NAME)
        if mustBePresent is not None:
            attributeSelector.mustBePresent = str2Bool(mustBePresent)
         
        return attributeSelector

                


"""NDG XACML equal function module - contains classes to represent XACML 1.0
*-equal functions

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "26/03/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: equal.py 7256 2010-07-29 20:29:29Z pjkersha $'
from ndg.xacml.core.context.exceptions import XacmlContextTypeError
from ndg.xacml.core.functions import AbstractFunction, FunctionClassFactoryBase


class EqualBase(AbstractFunction):
    """Generic equal function for all types
    
    @cvar TYPE: attribute type for the given implementation.  Derived classes
    should set appropriately
    @type TYPE: NoneType
    """
    TYPE = None
    
    def evaluate(self, attribute1, attribute2):
        """Match input attribute values
        
        @param attribute1: first of two attributes to match
        @type attribute1: ndg.xacml.core.attributevalue.AttributeValue derived
        @param attribute2: second attribute
        @type attribute2: ndg.xacml.core.attributevalue.AttributeValue derived
        @return: True if attributes match, False otherwise
        @rtype: bool
        """
        if not isinstance(attribute1, self.__class__.TYPE):
            raise XacmlContextTypeError('Expecting %r derived type for '
                                        '"attribute1"; got %r' %
                                        (self.__class__.TYPE, 
                                         type(attribute1)))
            
        if not isinstance(attribute2, self.__class__.TYPE):
            raise XacmlContextTypeError('Expecting %r derived type for '
                                        '"attribute2"; got %r' %
                                        (self.__class__.TYPE, 
                                         type(attribute2)))
            
        return attribute1.value == attribute2.value
    

class FunctionClassFactory(FunctionClassFactoryBase):
    """Class Factory for *-equal XACML function classes
    
    @cvar FUNCTION_NAMES: equal function URNs
    @type FUNCTION_NAMES: tuple
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for equal function URNs
    @type FUNCTION_NS_SUFFIX: string
    
    @cvar FUNCTION_BASE_CLASS: base class for all equal function classes
    @type FUNCTION_BASE_CLASS: ndg.xacml.core.functions.v1.EqualBase
    """
    FUNCTION_NAMES = (
        'urn:oasis:names:tc:xacml:1.0:function:string-equal',
        'urn:oasis:names:tc:xacml:1.0:function:boolean-equal',
        'urn:oasis:names:tc:xacml:1.0:function:integer-equal',
        'urn:oasis:names:tc:xacml:1.0:function:double-equal',
        'urn:oasis:names:tc:xacml:1.0:function:date-equal',
        'urn:oasis:names:tc:xacml:1.0:function:time-equal',
        'urn:oasis:names:tc:xacml:1.0:function:dateTime-equal',
        'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-equal',
        'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-equal',
        'urn:oasis:names:tc:xacml:1.0:function:anyURI-equal',
        'urn:oasis:names:tc:xacml:1.0:function:x500Name-equal',
        'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-equal',
        'urn:oasis:names:tc:xacml:1.0:function:hexBinary-equal',
        'urn:oasis:names:tc:xacml:1.0:function:base64Binary-equal',
        'urn:oasis:names:tc:xacml:1.0:function:xpath-node-equal'
    )
    FUNCTION_NS_SUFFIX = '-equal'
    FUNCTION_BASE_CLASS = EqualBase



"""NDG XACML Context Environment type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: environment.py 7087 2010-06-25 11:23:09Z pjkersha $"
from ndg.xacml.core.context import RequestChildBase


class Environment(RequestChildBase):
    """XACML Context Environment type"""
    ELEMENT_LOCAL_NAME = 'Environment'




#!/usr/bin/env python
"""NDG XACML functions unit test package 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "26/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: test_matchfunctions.py 7087 2010-06-25 11:23:09Z pjkersha $"
import unittest
from os import path
import logging
logging.basicConfig(level=logging.DEBUG)

from ndg.xacml.core.functions import FunctionMap
from ndg.xacml.core.functions.v2.regexp_match import RegexpMatchBase


class FunctionTestCase(unittest.TestCase):
    """Test XACML match functions implementation
    
    The log output gives an indication of the XACML functions which are not 
    implemented yet"""
    
    def test01LoadMap(self):   
        funcMap = FunctionMap()
        funcMap.loadAll()
        anyUriMatchNs = \
            'urn:oasis:names:tc:xacml:2.0:function:anyURI-regexp-match'
            
        self.assert_(issubclass(funcMap.get(anyUriMatchNs), RegexpMatchBase))

        
if __name__ == "__main__":
    unittest.main()


"""NDG XACML module for Response type 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "23/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: response.py 7087 2010-06-25 11:23:09Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.utils import TypedList
from ndg.xacml.core.context import XacmlContextBase
from ndg.xacml.core.context.result import Result


class Response(XacmlContextBase):
    """XACML Response type
    @cvar ELEMENT_LOCAL_NAME: XML local element name for the response
    @type ELEMENT_LOCAL_NAME: string

    @ivar __results: resource content
    @type __results: ndg.xacml.utils.TypedList
    """
    ELEMENT_LOCAL_NAME = 'Response'
    
    __slots__ = ('__results', )
    
    def __init__(self):
        """"Initialise results list"""
        super(Response, self).__init__()
        self.__results = TypedList(Result)
        
    @property
    def results(self):
        """Get Response results list
        
        @return: results list
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__results
    



"""NDG XACML Condition type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: condition.py 7101 2010-06-25 16:13:06Z pjkersha $"
from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.expression import Expression


class Condition(XacmlCoreBase):
    """XACML 2.0 Rule Condition Note the difference to XACML 1.0: the Condition 
    element is its own type and not an Apply type.  It expects a single 
    Expression derived type child element

    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string

    @cvar APPLY_ELEMENT_LOCAL_NAME: XML local name for the apply element
    @type APPLY_ELEMENT_LOCAL_NAME: string

    @ivar __expression: expression in this condition
    @type __expression: ndg.xacml.core.expression.Expression
    """
    ELEMENT_LOCAL_NAME = 'Condition'
    APPLY_ELEMENT_LOCAL_NAME = 'Apply'
    
    __slots__ = ('__expression', )
    
    def __init__(self):
        super(Condition, self).__init__()
        self.__expression = None
        
    @property
    def expression(self):
        """Get expression
        
        @return: expression for this condition
        @rtype: ndg.xacml.core.expression.Expression / NoneType
        """
        return self.__expression
        
    @expression.setter
    def expression(self, value):
        """Set expression
        
        @param value: expression for this condition
        @type value: ndg.xacml.core.expression.Expression
        @raise TypeError: incorrect input type set
        """
        if not isinstance(value, Expression):
            raise TypeError('Expecting Expression or Expression derived type '
                            'for "expression" attribute; got %r' %
                            type(value))
        self.__expression = value
    
    def evaluate(self, context):
        """Evaluate this rule condition
        @param context: the request context
        @type context: ndg.xacml.core.request.Request
        @return: True/False status for whether the rule condition matched or
        not
        @rtype: bool
        """
        result = self.expression.evaluate(context)
        
        return result


"""NDG XACML ElementTree reader module containing reader base class 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: reader.py 7109 2010-06-28 12:54:57Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from xml.etree import ElementTree

from ndg.xacml.core import XacmlPolicyBase
from ndg.xacml.parsers import AbstractReader


class ETreeAbstractReader(AbstractReader):
    """Base class for ElementTree implementation of XACML reader
    
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: NoneType"""
    TYPE = None
    
    def __init__(self):
        """Initialise ElementTree namespace map for XACML 2.0 namespaces
        
        @raise NotImplementedError: TYPE class variable must be set in a 
        derived class
        """
        if self.__class__.TYPE is None:
            raise NotImplementedError('No "TYPE" class variable set to specify '
                                      'the XACML type to instantiate')
            
        self.__namespace_map_backup = ElementTree._namespace_map.copy()
        ElementTree._namespace_map[''] = XacmlPolicyBase.XACML_2_0_POLICY_NS
        
    def __del__(self):
        """Restore original global namespace map"""
        ElementTree._namespace_map = self.__namespace_map_backup
    
    @staticmethod
    def _parse(obj):
        """Parse helper method
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: ElementTree element
        @rtype: xml.etree.Element
        """
        if ElementTree.iselement(obj):
            elem = obj
        else:
            # Treat as stream object
            elem = ElementTree.parse(obj).getroot()
            
        return elem
         
# Set up new class as an abstract base itself             
AbstractReader.register(ETreeAbstractReader)




"""NDG XACML AttributeDesignator type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: attributedesignator.py 7365 2010-08-25 15:41:20Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.utils import TypedList
from ndg.xacml.core.expression import Expression
from ndg.xacml.core.attributevalue import (AttributeValue, 
                                           AttributeValueClassFactory)
from ndg.xacml.core.context.request import Request
from ndg.xacml.core.context.handler import CtxHandlerInterface
from ndg.xacml.core.context.exceptions import MissingAttributeError


class AttributeDesignator(Expression):
    '''Base class for XACML Attribute Designator types
    
    @cvar ATTRIBUTE_ID_ATTRIB_NAME: attribute ID XML attribute name
    @type ATTRIBUTE_ID_ATTRIB_NAME: string
    @cvar ISSUER_ATTRIB_NAME: issuer XML attribute name
    @type ISSUER_ATTRIB_NAME: string
    @cvar MUST_BE_PRESENT_ATTRIB_NAME: must be present XML attribute name
    @type MUST_BE_PRESENT_ATTRIB_NAME: string
    
    @ivar __attributeId: attribute ID for this designator
    @type __attributeId: basestring / NoneType
    @ivar __issuer: issuer if the designator
    @type __issuer: basestring / NoneType
    @ivar __mustBePresent: XML must be present flag
    @type __mustBePresent: bool
    @ivar __attributeValueFactory: When evaluating matches, use an attribute 
    value class factory to create attribute values for match bag of the correct 
    DataType to respect type based rule functions
    @type __attributeValueFactory: ndg.xacml.core.attributevalue.AttributeValueClassFactory
    '''
    ATTRIBUTE_ID_ATTRIB_NAME = 'AttributeId'
    ISSUER_ATTRIB_NAME = 'Issuer'
    MUST_BE_PRESENT_ATTRIB_NAME = 'MustBePresent'
    
    __slots__ = (
        '__attributeId', 
        '__issuer', 
        '__mustBePresent',
        '__attributeValueFactory'
    )
    
    def __init__(self):
        """Initialise attributes"""
        super(AttributeDesignator, self).__init__()
        self.__attributeId = None
        self.__issuer = None
        self.__mustBePresent = False
        
        # When evaluating matches, use an attribute value class factory to 
        # create attribute values for match bag of the correct DataType to 
        # respect type based rule functions
        self.__attributeValueFactory = AttributeValueClassFactory()

    @property
    def attributeId(self):
        """Get Attribute Id
        @return: attribute ID
        @rtype: basestring / NoneType
        """
        return self.__attributeId

    @attributeId.setter
    def attributeId(self, value):
        """Set Attribute Id
        @param value: attribute ID
        @type value: basestring 
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "attributeId" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__attributeId = value  
        
    @property
    def issuer(self):
        """Get Issuer
        @return: issuer
        @rtype: basestring / NoneType
        """
        return self.__issuer

    @issuer.setter
    def issuer(self, value):
        """Set Issuer
        @param value: issuer
        @type value: basestring
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "issuer" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__issuer = value   

    @property
    def mustBePresent(self):
        """Get Must Be Present flag
        @return: must be present flag
        @rtype: bool
        """
        return self.__mustBePresent

    @mustBePresent.setter
    def mustBePresent(self, value):
        """Set Must Be Present flag
        @param value: must be present flag
        @type value: bool
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, bool):
            raise TypeError('Expecting %r type for "mustBePresent" '
                            'attribute; got %r' % (bool, type(value)))
            
        self.__mustBePresent = value   
             
    @property
    def attributeValueFactory(self):
        """Get Attribute Value factory function
        
        @return: attribute value factory instance
        @rtype: ndg.xacml.core.attributevalue.AttributeValueClassFactory
        """
        return self.__attributeValueFactory 
                           
        
class SubjectAttributeDesignator(AttributeDesignator):
    """XACML Subject Attribute Designator type
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    """
    ELEMENT_LOCAL_NAME = 'SubjectAttributeDesignator'
    __slots__ = ()
    
    def evaluate(self, context):
        """Evaluate the result of the SubjectAttributeDesignator in a condition
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: attribute value(s) resulting from execution of this expression
        in a condition
        @rtype: AttributeValue/NoneType  
        """ 
        if not isinstance(context, Request):
            raise TypeError('Expecting %r type for context input; got %r' %
                            (Request, type(context)))
        
        dataType = self.dataType
        attributeValueBag = TypedList(self.attributeValueFactory(dataType))
        attributeId = self.attributeId
        issuer = self.issuer
        
        log.debug("In SubjectAttributeDesignator for attribute %r", attributeId)
        
        if issuer is not None:
            _issuerMatch = lambda _issuer: issuer == _issuer
        else:
            _issuerMatch = lambda _issuer: True
        
        _attributeMatch = lambda attr: (attr.attributeId == attributeId and 
                                        attr.dataType == dataType and
                                        _issuerMatch(attr.issuer))
                                        
        for subject in context.subjects:  
            for attr in subject.attributes:
                if _attributeMatch(attr):
                    attributeValueBag.extend([i for i in attr.attributeValues
                                              if i.dataType == dataType])
                    
            if context.ctxHandler is not None:
                # Try querying the Policy Information Point via the Context 
                # Handler to see if values for the attribute specified in this 
                # designator can be retrieved externally. If retrieved, they're 
                # added to the bag.
                log.debug("Making query to PIP for additional subject "
                          "attributes to satify match")
                
                attributeValues = context.ctxHandler.pipQuery(context, self)
                if attributeValues is not None:
                    log.debug("PIP retrieved additional subject attributes: %r",
                              attributeValues)
                    
                    # Weed out any duplicates
                    if len(attributeValueBag) > 0:
                        filtAttributeValues = [attrVal 
                                           for attrVal in attributeValues 
                                           if attrVal not in attributeValueBag]
                    else:
                        filtAttributeValues = attributeValues
                        
                    attributeValueBag.extend(filtAttributeValues)
                    
        if len(attributeValueBag) == 0 and self.mustBePresent:
            raise MissingAttributeError('"MustBePresent" is set for %r but no '
                                        'match for attributeId=%r, dataType=%r '
                                        'and issuer=%r' % 
                                        (self.__class__.__name__,
                                         self.attributeId,
                                         self.dataType,
                                         self.issuer)) 
                        
        return attributeValueBag
        
        
class ResourceAttributeDesignator(AttributeDesignator):
    """XACML Resource Attribute Designator type
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    """
    ELEMENT_LOCAL_NAME = 'ResourceAttributeDesignator'
    __slots__ = ()
    
    def evaluate(self, context):
        """Evaluate the result of the ResourceAttributeDesignator in a condition
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: attribute value(s) resulting from execution of this expression
        in a condition
        @rtype: AttributeValue/NoneType  
        """ 
        if not isinstance(context, Request):
            raise TypeError('Expecting %r type for context input; got %r' %
                            (Request, type(context)))
        
        dataType = self.dataType
        attributeValueBag = TypedList(self.attributeValueFactory(dataType))
        attributeId = self.attributeId
        issuer = self.issuer
        
        if issuer is not None:
            _issuerMatch = lambda _issuer: issuer == _issuer
        else:
            _issuerMatch = lambda _issuer: True
        
        _attributeMatch = lambda attr: (attr.attributeId == attributeId and 
                                        attr.dataType == dataType and
                                        _issuerMatch(attr.issuer))
                    
        for resource in context.resources:
            for attr in resource.attributes:
                if _attributeMatch(attr):
                    attributeValueBag.extend([i for i in attr.attributeValues
                                              if i.dataType == dataType])
                    
        if len(attributeValueBag) == 0 and self.mustBePresent:
            raise MissingAttributeError('"MustBePresent" is set for %r but no '
                                        'match for attributeId=%r, dataType=%r '
                                        'and issuer=%r' % 
                                        (self.__class__.__name__,
                                         self.attributeId,
                                         self.dataType,
                                         self.issuer)) 
                        
        return attributeValueBag
    
        
class ActionAttributeDesignator(AttributeDesignator):
    """XACML Action Attribute Designator type
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    """
    ELEMENT_LOCAL_NAME = 'ActionAttributeDesignator'
    __slots__ = ()
    
    def evaluate(self, context):
        """Evaluate the result of the ActionAttributeDesignator in a condition
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: attribute value(s) resulting from execution of this expression
        in a condition
        @rtype: AttributeValue/NoneType  
        """ 
        if not isinstance(context, Request):
            raise TypeError('Expecting %r type for context input; got %r' %
                            (Request, type(context)))
        
        dataType = self.dataType
        attributeValueBag = TypedList(self.attributeValueFactory(dataType))
        attributeId = self.attributeId
        issuer = self.issuer
        action = context.action
        
        if issuer is not None:
            _issuerMatch = lambda _issuer: issuer == _issuer
        else:
            _issuerMatch = lambda _issuer: True
        
        _attributeMatch = lambda attr: (attr.attributeId == attributeId and 
                                        attr.dataType == dataType and
                                        _issuerMatch(attr.issuer))
                    
        for attr in action.attributes:
            if _attributeMatch(attr):
                attributeValueBag.extend([i for i in attr.attributeValues
                                          if i.dataType == dataType])
                    
        if len(attributeValueBag) == 0 and self.mustBePresent:
            raise MissingAttributeError('"MustBePresent" is set for %r but no '
                                        'match for attributeId=%r, dataType=%r '
                                        'and issuer=%r' % 
                                        (self.__class__.__name__,
                                         self.attributeId,
                                         self.dataType,
                                         self.issuer)) 
                        
        return attributeValueBag    
    
    
class EnvironmentAttributeDesignator(AttributeDesignator):
    """XACML Environment Attribute Designator type
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    """
    ELEMENT_LOCAL_NAME = 'EnvironmentAttributeDesignator'
    __slots__ = ()
    
    def evaluate(self, context):
        """Evaluate the result of the EnvironmentAttributeDesignator in a 
        condition
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: attribute value(s) resulting from execution of this expression
        in a condition
        @rtype: AttributeValue/NoneType  
        """ 
        if not isinstance(context, Request):
            raise TypeError('Expecting %r type for context input; got %r' %
                            (Request, type(context)))
        
        dataType = self.dataType
        attributeValueBag = TypedList(self.attributeValueFactory(dataType))
        attributeId = self.attributeId
        issuer = self.issuer
        environment = context.environment
        
        if issuer is not None:
            _issuerMatch = lambda _issuer: issuer == _issuer
        else:
            _issuerMatch = lambda _issuer: True   
        
        _attributeMatch = lambda attr: (attr.attributeId == attributeId and 
                                        attr.dataType == dataType and
                                        _issuerMatch(attr.issuer))
                 
        for attr in environment.attributes:
            if _attributeMatch(attr):
                attributeValueBag.extend([i for i in attr.attributeValues
                                          if i.dataType == dataType])
                    
        if len(attributeValueBag) == 0 and self.mustBePresent:
            raise MissingAttributeError('"MustBePresent" is set for %r but no '
                                        'match for attributeId=%r, dataType=%r '
                                        'and issuer=%r' % 
                                        (self.__class__.__name__,
                                         self.attributeId,
                                         self.dataType,
                                         self.issuer)) 
                        
        return attributeValueBag


"""NDG Security Match type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: match.py 7351 2010-08-20 14:46:35Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.attributevalue import AttributeValue
from ndg.xacml.core.attributedesignator import AttributeDesignator
from ndg.xacml.core.attributeselector import AttributeSelector
from ndg.xacml.core.functions import (FunctionMap, UnsupportedStdFunctionError,
                                      UnsupportedFunctionError)
from ndg.xacml.core.context.exceptions import XacmlContextError


class MatchBase(XacmlCoreBase):
    """Base class for representation of SubjectMatch, ResourceMatch, 
    ActionMatch and EnvironmentMatch Target elements
    
    @cvar ELEMENT_LOCAL_NAME: XML Local Name of this element 
    @type ELEMENT_LOCAL_NAME: string
   
    @cvar MATCH_ID_ATTRIB_NAME: XML attribute name for match ID
    @type MATCH_ID_ATTRIB_NAME: string
   
    @cvar ATTRIBUTE_VALUE_ELEMENT_LOCAL_NAME: XML Local Name of attribute value
    child element 
    @type ATTRIBUTE_VALUE_ELEMENT_LOCAL_NAME: string

    @ivar __attributeValue: attribute value associated with this match
    @type __attributeValue: ndg.xacml.core.attributevalue.AttributeValue
    @ivar __attributeDesignator: attribute designator - only a designator or 
    selector may be set for a given instance not both
    @type __attributeDesignator: ndg.xacml.core.attributedesignator.AttributeDesignator
    @ivar __attributeSelector: attribute selector - only a designator or 
    selector may be set for a given instance not both
    @type __attributeSelector: ndg.xacml.core.attributeselector.AttributeSelector
    @ivar __matchId: match identifier
    @type __matchId: NoneType / basestring
    @ivar __function: function to be applied
    @type __function: ndg.xacml.core.functions.AbstractFunction derived type
    @ivar __functionMap: function mapping object to map URNs to function class
    implementations
    @type __functionMap: ndg.xacml.core.functions.FunctionMap
    @ivar __loadFunctionFromId: boolean determines whether or not to load
    function classes for given function URN in functionId set property method
    @type __loadFunctionFromId: bool
    """
    ELEMENT_LOCAL_NAME = None
    MATCH_ID_ATTRIB_NAME = 'MatchId'
    ATTRIBUTE_VALUE_ELEMENT_LOCAL_NAME = 'AttributeValue'

    __slots__ = (
        '__attributeValue', 
        '__attributeDesignator', 
        '__attributeSelector',
        '__matchId',
        '__function', 
        '__functionMap',
        '__loadFunctionFromId',
    )
    
    def __init__(self):
        """Initial attributes corresponding to the equivalent XACML schema type
        and also create a function map to map functions from MatchIds
        """
        self.__attributeValue = None
         
        # Either/or in schema
        self.__attributeDesignator = None
        self.__attributeSelector = None
        
        self.__matchId = None
        
        self.__function = None
        self.__functionMap = FunctionMap()
        self.__loadFunctionFromId = True
        
    @property
    def attributeValue(self):
        """Match attribute value
        
        @return: attribute value
        @rtype: ndg.xacml.core.attributevalue.Attribute"""
        return self.__attributeValue
    
    @attributeValue.setter
    def attributeValue(self, value):
        """Set match attribute value.
        @param value: attribute value
        @type value: ndg.xacml.core.attributevalue.AttributeValue
        @raise TypeError: incorrect type set
        """
        if not isinstance(value, AttributeValue):
            raise TypeError('Expecting %r type for "matchId" '
                            'attribute; got %r' % 
                            (AttributeValue, type(value)))
            
        self.__attributeValue = value
        
    @property
    def attributeDesignator(self):
        """@return: attribute designator - only a designator or 
        selector may be set for a given instance not both
        @rtype: ndg.xacml.core.attributedesignator.AttributeDesignator
        """
        return self.__attributeDesignator
    
    @attributeDesignator.setter
    def attributeDesignator(self, value):
        """Set match attribute designator.  Match may have an 
        attributeDesignator or an attributeSelector setting a designator DELETES
        any attributeSelector previously set
        
        @param value: attribute selector - only a designator or 
        selector may be set for a given instance not both
        @type value: ndg.xacml.core.attributeselector.AttributeSelector  
        @raise TypeError: incorrect type for input value      
        """
        if not isinstance(value, AttributeDesignator):
            raise TypeError('Expecting %r type for "attributeDesignator" '
                            'attribute; got %r' % 
                            (AttributeDesignator, type(value)))
            
        self.__attributeDesignator = value
        self.__attributeSelector = None
 
    @property
    def attributeSelector(self):
        '''
        @return: attribute selector
        @rtype: ndg.xacml.core.attributeselector.AttributeSelector
        '''
        return self.__attributeSelector
    
    @attributeSelector.setter
    def attributeSelector(self, value):
        """Set match attribute selector.  Match may have an 
        attributeDesignator or an attributeSelector setting a selector DELETES
        any attributeDesignator previously set
        
        @param value: attribute selector
        @type value: ndg.xacml.core.attributeselector.AttributeSelector
        """
        if not isinstance(value, AttributeSelector):
            raise TypeError('Expecting %r type for "matchId" '
                            'attribute; got %r' % 
                            (AttributeSelector, type(value)))
            
        self.__attributeSelector = value
        self.__attributeDesignator = None
                       
    def _getMatchId(self):
        """Match identifier for match function
        @return: match identifier
        @rtype: NoneType / basestring
        """
        return self.__matchId

    def _setMatchId(self, value):
        """Match identifier for match function
        @param value: match identifier
        @type value: basestring
        @raise TypeError: if incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "matchId" '
                            'attribute; got %r' % type(value))
            
        self.__matchId = value
        
        # Also retrieve function for this match ID if a map has been set
        if self.__loadFunctionFromId:
            self.setFunctionFromMap(self.__functionMap)   

    matchId = property(_getMatchId, _setMatchId, None, 
                       "Match identifier for match function")
      
    @property
    def loadFunctionFromId(self):
        """Set to False to stop the functionId property set method automatically
        trying to load the corresponding function for the given functionId 

        @return: boolean determines whether or not to load
        function classes for given function URN in functionId set property 
        method
        @rtype: bool
        """
        return self.__loadFunctionFromId
    
    @loadFunctionFromId.setter
    def loadFunctionFromId(self, value):
        """
        @param value: boolean determines whether or not to load
        function classes for given function URN in functionId set property 
        method
        @type value: bool
        """
        if not isinstance(value, bool):
            raise TypeError('Expecting %r type for "loadFunctionFromId" '
                            'attribute; got %r' % (bool, type(value)))
            
        self.__loadFunctionFromId = value
        
    def setFunctionFromMap(self, functionMap):
        """Set the function from a function map - a dictionary of function ID to
        function mappings.  The function is looked up based on the "functionId"
        attribute.  This method is automatically called when the functionId set
        property method is invoked.  To switch off this behaviour set
        
        loadFunctionFromId = False
        
        @param functionMap: mapping of function URNs to function classes
        @type functionMap: dict like object
        @raise UnsupportedStdFunctionError: policy references a function type 
        which is in the XACML spec. but is not supported by this implementation
        @raise UnsupportedFunctionError: policy references a function type which
        is not supported by this implementation
        """
        if self.matchId is None:
            raise AttributeError('"functionId" attribute must be set in order '
                                 'to retrieve the required function')
            
        # Get function class for this <Apply> statement         
        functionClass = functionMap.get(self.matchId)
        if functionClass is NotImplemented:
            raise UnsupportedStdFunctionError('No match function class '
                                              'implemented for MatchId="%s"' % 
                                              self.matchId)
        elif functionClass is None:
            raise UnsupportedFunctionError('<Apply> function namespace %r is '
                                           'not recognised' % 
                                           self.matchId) 
            
        self.__function = functionClass()
    
    @property
    def functionMap(self):
        """functionMap object for PDP to retrieve functions from given XACML
        function URNs
        @return: function mapping object to map URNs to function 
        class implementations
        @rtype: ndg.xacml.core.functions.FunctionMap
        """
        return self.__functionMap
    
    @functionMap.setter
    def functionMap(self, value):
        '''functionMap object for PDP to retrieve functions from given XACML
        function URNs
        
        @param value: function mapping object to map URNs to function class
        implementations
        @type value: ndg.xacml.core.functions.FunctionMap
        @raise TypeError: raise if input value is incorrect type
        '''
        if not isinstance(value, FunctionMap):
            raise TypeError('Expecting %r derived type for "functionMap" '
                            'input; got %r instead' % (FunctionMap, 
                                                       type(value)))
        self.__functionMap = value
          
    @property  
    def function(self):
        """Function for this <Apply> instance
        @return: function to be applied
        @rtype: ndg.xacml.core.functions.AbstractFunction derived type
        """
        return self.__function
        
    def evaluate(self, context):
        """Evaluate the match object against the relevant element in the request
        context
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: match status
        @rtype: bool
        """
        
        # Create a match function based on the presence or absence of an
        # AttributeDesignator or AttributeSelector
        if self.attributeDesignator is not None:
            requestAttributeValues = self.attributeDesignator.evaluate(context)
            
        elif self.attributeSelector is not None:
            # Nb. Evaluation is not currently supported.  This will require that
            # the request provide a reference to it's XML representation and an 
            # abstraction of the XML parser for executing XPath searches into 
            # that representation
            requestAttributeValues = self.attributeSelector.evaluate(context)
        else:
            raise XacmlContextError('No attribute designator or selector set '
                                    'for Target Match element %r with MatchId '
                                    '= %r and attributeValue = %r' %
                                    (self.__class__.ELEMENT_LOCAL_NAME,
                                     self.matchId,
                                     self.attributeValue))
            
        # Iterate through each attribute in the request in turn matching it
        # against the target using the generated _attributeMatch function
        #
        # Any Match element NOT matching will result in an overall status of
        # no match.
        #
        # Continue iterating through the whole list even if a False status
        # is found.  The other attributes need to be checked in case an 
        # error occurs.  In this case the top-level PDP exception handling 
        # block will catch it and set an overall decision of INDETERMINATE
        attrMatchStatusValues = [False]*len(requestAttributeValues)
        matchFunction = self.function
        matchAttributeValue = self.attributeValue
        
        for i, requestAttributeValue in enumerate(requestAttributeValues):
            
            attrMatchStatusValues[i] = matchFunction.evaluate(
                                                         matchAttributeValue,
                                                         requestAttributeValue)
            if attrMatchStatusValues[i] == True:
                if log.getEffectiveLevel() <= logging.DEBUG:
                    log.debug('Target attribute value %r matches request '
                              'attribute value %r matches using match '
                              'function Id %r',
                              matchAttributeValue,
                              requestAttributeValue,
                              self.matchId)
        
        # Need check for len() because all([]) yields True!           
        matchStatus = (len(attrMatchStatusValues) > 0 and 
                       all(attrMatchStatusValues))
        
        return matchStatus
    
    
class SubjectMatch(MatchBase):
    "Subject Match Type"
    ELEMENT_LOCAL_NAME = 'SubjectMatch'
    ATTRIBUTE_DESIGNATOR_ELEMENT_LOCAL_NAME = 'SubjectAttributeDesignator'   
    
    
class ResourceMatch(MatchBase):
    "Resource Match"
    ELEMENT_LOCAL_NAME = 'ResourceMatch'
    ATTRIBUTE_DESIGNATOR_ELEMENT_LOCAL_NAME = 'ResourceAttributeDesignator'
    
    
class ActionMatch(MatchBase):
    "Action match"
    ELEMENT_LOCAL_NAME = 'ActionMatch'
    ATTRIBUTE_DESIGNATOR_ELEMENT_LOCAL_NAME = 'ActionAttributeDesignator'    
    
    
class EnvironmentMatch(MatchBase):
    "Environment Match"
    ELEMENT_LOCAL_NAME = 'EnvironmentMatch'
    ATTRIBUTE_DESIGNATOR_ELEMENT_LOCAL_NAME = 'EnvironmentAttributeDesignator'


#!/usr/bin/env python

"""NDG XACML

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: setup.py 7445 2010-09-03 13:55:01Z pjkersha $'

# Bootstrap setuptools if necessary.
from ez_setup import use_setuptools
use_setuptools()
from setuptools import setup, find_packages

_longDescription = """\
XACML 2.0 implementation for CEDA (the Centre for Environmental Data Archival) 
STFC, Rutherford Appleton Laboratory.  This is follow on work from the NERC 
(Natural Environment Research Council) DataGrid 3 Project.

XACML (eXtensible Access Control Mark-up Language), is an XML based language for
expressing access control policies.

See: http://www.oasis-open.org/committees/xacml/

Release 0.3
===========
Includes important fixes for equals functions, and improvement to at least one
member functions.  Unit tests improved with wider coverage of different rule
definitions and example request contexts.

Improved and added to support for context handler and Policy Information Point
interfaces including the ability for the PDP to call back to a PIP via a 
Context handler to retrieve additional subject attributes.

Release 0.2
===========
Only the parts of the specification immediately required for CEDA have been 
implemented in this initial release:
 Policy Decision Point;
 Deny overrides and Permit overrides rule combining algorithms;
 AttributeDesignators;
 various function types: see ndg.xacml.core.functions;
 and attribute types: see ndg.xacml.core.attribute;
 incomplete support for <AttributeSelector>s, <VariableReference>, 
 <VariableDefinition>. <Obligations>;
 includes an ElementTree based parser for Policies. No support for writing
 out policies or read/write of XML representation of <Request> and <Response>;
   
See ndg.xacml.test for unit tests and examples.

The software follows a modular structure to allow it to be extended easily to 
include new parsers, functions and attribute types 

"""

setup(
    name =           		'ndg_xacml',
    version =        		'0.3',
    description =           'XACML 2.0 implementation for the NERC DataGrid',
    long_description =		_longDescription,
    author =         		'Philip Kershaw',
    author_email =   		'Philip.Kershaw@stfc.ac.uk',
    maintainer =         	'Philip Kershaw',
    maintainer_email =   	'Philip.Kershaw@stfc.ac.uk',
    url =            		'http://proj.badc.rl.ac.uk/ndg/wiki/Security/XACML',
    license =               'BSD - See LICENCE file for details',
#    install_requires =		[],
    dependency_links =		["http://ndg.nerc.ac.uk/dist"],
    packages =       		find_packages(),
    namespace_packages =	['ndg'],
    package_data =		    {
        'ndg.xacml.core': ['documentation/Makefile'],
        'ndg.xacml.test': ['*.xml'],
    },
    entry_points =          None,
    test_suite =		    'ndg.xacml.test',
    zip_safe =              False,
    classifiers =           [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Security',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



"""NDG XACML one and only functions module

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "01/04/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: $'
from ndg.xacml.core.functions import (AbstractFunction, 
                                      FunctionClassFactoryInterface)
from ndg.xacml.core.context.exceptions import XacmlContextTypeError
from ndg.xacml.utils import TypedList as Bag


class And(AbstractFunction):
    """Base class for XACML <type>-and functions
    
    @cvar FUNCTION_NS: namespace for this function
    @type FUNCTION_NS: string
    @cvar BAG_TYPE: type for 
    @type BAG_TYPE:
    """
    FUNCTION_NS = AbstractFunction.V1_0_FUNCTION_NS + 'and'
    BAG_TYPE = bool
    
    def evaluate(self, bag):
        """perform AND function on the elements in the bag ref.
        
        access_control-xacml-2.0-core-spec-os, Fe 2005 - A.3.5 Logical functions
        
        @param bag: bag containing elements to be AND'ed
        @type bag: ndg.xacml.utils.TypedList
        
        @return: result of AND operation on the inputs
        @rtype: bool
        """
        if not isinstance(bag, Bag):
            raise XacmlContextTypeError('Expecting %r derived type for "bag"; '
                                    'got %r' % (Bag, type(bag)))
            
        if bag.elementType != self.__class__.BAG_TYPE:
            raise XacmlContextTypeError('Expecting %r type elements for "bag"; '
                                        'got %r' %
                                        (self.__class__.BAG_TYPE, 
                                         bag.elementType))
                        
        nBagElems = len(bag)
        if nBagElems == 0:
            return True
        else:
            for elem in bag:
                if not elem:
                    return False
            
        return True

    
class FunctionClassFactory(FunctionClassFactoryInterface):
    """Class Factory for and XACML function class
    
    @cvar FUNCTION_NS: URN for and function
    @type FUNCTION_NS: string
    """
    FUNCTION_NS = 'urn:oasis:names:tc:xacml:1.0:function:and'
    
    def __call__(self, identifier):
        '''Create class for the And XACML function identifier
        
        @param identifier: XACML and function identifier
        @type identifier: basestring
        @return: and function class or None if identifier doesn't match
        @rtype: ndg.xacml.core.functions.v1.and.And / NoneType
        '''
        if identifier == And.FUNCTION_NS:
            return And
        else:
            return None



"""NDG XACML AttributeSelector type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: attributeselector.py 7101 2010-06-25 16:13:06Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core.expression import Expression


class AttributeSelector(Expression):
    '''XACML Attribute Selector type 
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string    
    @cvar REQUEST_CONTEXT_PATH_ATTRIB_NAME: request context XML attribute name
    @type REQUEST_CONTEXT_PATH_ATTRIB_NAME: string
    @cvar MUST_BE_PRESENT_ATTRIB_NAME: must be present XML attribute name
    @type MUST_BE_PRESENT_ATTRIB_NAME: string

    '''
    ELEMENT_LOCAL_NAME = 'AttributeSelector'
    MUST_BE_PRESENT_ATTRIB_NAME = 'MustBePresent'
    REQUEST_CONTEXT_PATH_ATTRIB_NAME = 'RequestContextPath'
    
    __slots__ = ('__mustBePresent', '__requestContextPath')

    def __init__(self):
        '''XACML Attribute Selector type 
        '''
        super(AttributeSelector, self).__init__()
        self.__mustBePresent = None
        self.__requestContextPath = None
        
    @property
    def requestContextPath(self):
        """Get request context path
        @return: request context path
        @rtype: basestring
        """
        return self.__requestContextPath

    @requestContextPath.setter
    def requestContextPath(self, value):
        """Set request context path
        @param value: request context path
        @type value: basestring
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "requestContextPath" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__requestContextPath = value   

    @property
    def mustBePresent(self):
        """Get Must Be Present flag
        @return: must be present flag
        @rtype: bool
        """
        return self.__mustBePresent

    @mustBePresent.setter
    def mustBePresent(self, value):
        """Set Must Be Present flag
        @param value: must be present flag
        @type value: bool
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, bool):
            raise TypeError('Expecting %r type for "mustBePresent" '
                            'attribute; got %r' % (bool, type(value)))
            
        self.__mustBePresent = value 

    def evaluate(self, context):
        """Placeholder to override abstract method and enable this class to be
        instantiated.  However, AttributeSelectors are not fully implemented so
        this method raises a not implemented error.
        """
        raise NotImplementedError('PDP evaluation for AttributeSelectors is '
                                  'not currently implemented')


"""NDG XACML attribute type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: attributevalue.py 7412 2010-09-02 08:25:48Z pjkersha $"
from datetime import datetime, timedelta

from ndg.xacml.utils import VettedDict
from ndg.xacml.core.expression import Expression
    

class AttributeValue(Expression):
    """XACML Attribute Value type
    
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string  
    @cvar CLASS_NAME_SUFFIX: all attribute value classes end with this suffix
    @type CLASS_NAME_SUFFIX: string
    @cvar IDENTIFIER_PREFIX: geberic prefix for attribute value URNs
    @type IDENTIFIER_PREFIX: string
    @cvar IDENTIFIER: URN for attribute value in derived class 
    @type IDENTIFIER: NoneType - derived classes should set to appropriate 
    string
    @cvar TYPE_URIS: URIs for all the different supported types
    @type TYPE_URIS: tuple
    @cvar TYPE_NAMES: corresponding short names for all the types
    @type TYPE_NAMES: tuple
    @cvar NATIVE_TYPES: equivalent python types as implemented
    @cvar TYPE_MAP: mapping from type names to python types
    @type TYPE_MAP: dict
    @cvar TYPE_URI_MAP: mapping from type names to type URIs
    @type TYPE_URI_MAP: dict
    @cvar TYPE: type name for derived type - set to None in this parent class
    @type TYPE: NoneType / string in derived type
    
    @ivar __value: setting for this attribute value
    @type __value: any - constrained in derived classes
    """
    ELEMENT_LOCAL_NAME = 'AttributeValue'
    CLASS_NAME_SUFFIX = 'AttributeValue'
    IDENTIFIER_PREFIX = 'http://www.w3.org/2001/XMLSchema#'
  
    IDENTIFIER = None
    
    STRING_TYPE_URI = IDENTIFIER_PREFIX + 'string'
    ANY_TYPE_URI = IDENTIFIER_PREFIX + 'anyURI'
    INTEGER_TYPE_URI = IDENTIFIER_PREFIX + 'integer'
    BOOLEAN_TYPE_URI = IDENTIFIER_PREFIX + 'boolean'
    DOUBLE_TYPE_URI = IDENTIFIER_PREFIX + 'double'
    DATE_TYPE_URI = IDENTIFIER_PREFIX + 'date'
    DATETIME_TYPE_URI = IDENTIFIER_PREFIX + 'dateTime'
    TIME_TYPE_URI = IDENTIFIER_PREFIX + 'time'
    DAYTIMEDURATION_TYPE_URI = \
    'http://www.w3.org/TR/2002/WD-xquery-operators-20020816#dayTimeDuration'
    YEARMONTHDURATION_TYPE_URI = \
    'http://www.w3.org/TR/2002/WD-xquery-operators-20020816#yearMonthDuration'
    X500NAME_TYPE_URI = 'urn:oasis:names:tc:xacml:1.0:data-type:x500Name'
    RFC822NAME_TYPE_URI = 'urn:oasis:names:tc:xacml:1.0:data-type:rfc822Name'
    HEXBINARY_TYPE_URI = IDENTIFIER_PREFIX + 'hexBinary'
    BASE64BINARY_TYPE_URI = IDENTIFIER_PREFIX + 'base64Binary'
    IPADDRESS_TYPE_URI = 'urn:oasis:names:tc:xacml:2.0:data-type:ipAddress'
    DNSNAME_TYPE_URI = 'urn:oasis:names:tc:xacml:2.0:data-type:dnsName'

    TYPE_URIS = (
        STRING_TYPE_URI,
        ANY_TYPE_URI,
        INTEGER_TYPE_URI,
        BOOLEAN_TYPE_URI,
        DOUBLE_TYPE_URI,
        DATE_TYPE_URI,
        DATETIME_TYPE_URI,
        TIME_TYPE_URI,
        DAYTIMEDURATION_TYPE_URI,
        YEARMONTHDURATION_TYPE_URI,
        X500NAME_TYPE_URI,
        RFC822NAME_TYPE_URI,
        HEXBINARY_TYPE_URI,
        BASE64BINARY_TYPE_URI,
        IPADDRESS_TYPE_URI,
        DNSNAME_TYPE_URI
    )
    TYPE_NAMES = (
        'String',
        'AnyURI',
        'Integer',
        'Boolean',
        'Double',
        'Date',
        'DateTime',
        'Time',
        'DayTimeDuration',
        'YearMonthDuration',
        'X500Name',
        'Rfc822Name',
        'HexBinary',
        'Base64Binary',
        'IpAddress',
        'DnsName',
    )
    NATIVE_TYPES = (
        basestring,
        basestring,
        int,
        bool,
        float,
        datetime,
        datetime,
        datetime,
        timedelta,
        timedelta,
        basestring,
        basestring,
        int,
        NotImplemented,
        basestring,
        basestring
    )
    TYPE_MAP = dict(zip(TYPE_NAMES, NATIVE_TYPES))
    TYPE_URI_MAP = dict(zip(TYPE_NAMES, TYPE_URIS))
    TYPE = None
    
    __slots__ = ('__value',) 
    
    def __init__(self, value=None):
        """Derived classes must override setting TYPE class variable
        
        @param value: initialise the attribute value by setting this keyword
        @type value: (set by self.__class__.TYPE)
        """
        
        super(AttributeValue, self).__init__()
        if self.__class__.TYPE is None:
            raise NotImplementedError('TYPE class variable must be set to a '
                                      'valid type in a derived class')
            
        self.__value = None
        
        # Allow derived classes to make an implicit data type setting
        self.dataType = self.__class__.IDENTIFIER
        
        if value is not None:
            self.value = value

    def __repr__(self):
        return "%s = %r " % (super(AttributeValue, self).__repr__(),
                             self.__value)

    def __eq__(self, attrVal):
        """Check for equality by comparing the value attributes"""
        if not isinstance(attrVal, self.__class__):
            raise TypeError('Expecting %r type for "value" '
                            'attribute; got %r' % (self.__class__.TYPE, 
                                                   type(attrVal)))
            
        return self.__value == attrVal.value
    
    def _get_value(self):
        """Get value
        @return: setting for this attribute value
        @rtype: any - constrained in derived classes
        """
        return self.__value

    def _set_value(self, value):
        """Set value
        
        @param value: setting for this attribute value
        @type value: any - constrained in derived classes
        @raise TypeError: if type doesn't match TYPE class variable.  Derived
        classes should set this
        """
        if not isinstance(value, self.__class__.TYPE):
            raise TypeError('Expecting %r type for "value" '
                            'attribute; got %r' % (self.__class__.TYPE, 
                                                   type(value)))
            
        self.__value = value  

    value = property(_get_value, _set_value, None, "expression value") 
    
    def evaluate(self, context):
        """Evaluate the result of the expression in a condition.  In the case of
        an attribute value it's simply itself
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: this attribute value
        @rtype: AttributeValue  
        """ 
        return self


class AttributeValueClassMap(VettedDict):
    """Specialised dictionary to hold mappings of XML attribute type URIs to 
    their equivalent classes
    """
    
    def __init__(self):
        """Force entries to derive from AttributeValue and IDs to
        be string type
        """        
        # Filters are defined as staticmethods but reference via self here to 
        # enable derived class to override them as standard methods without
        # needing to redefine this __init__ method            
        VettedDict.__init__(self, self.keyFilter, self.valueFilter)
        
    @staticmethod
    def keyFilter(key):
        """Enforce string type keys
        
        @param key: URN for attribute
        @type key: basestring
        @return: boolean True indicating key is OK
        @rtype: bool
        @raise TypeError: incorrect input type
        """
        if not isinstance(key, basestring):
            raise TypeError('Expecting %r type for key; got %r' % 
                            (basestring, type(key))) 
        return True 
    
    @staticmethod
    def valueFilter(value):
        """Enforce AttributeValue derived types for values
        @param value: attribute value
        @type value: ndg.xacml.core.attributevalue.AttributeValue derived type
        @return: boolean True indicating attribute value is correct type
        @rtype: bool
        @raise TypeError: incorrect input type
        """
        if not issubclass(value, AttributeValue):
            raise TypeError('Expecting %r derived type for value; got %r' % 
                            (AttributeValue, type(value))) 
        return True 


# Dynamically Create classes based on AttributeValue for all the XACML primitive
# types
_IDENTIFIER2CLASS_MAP = AttributeValueClassMap()

for typeName, _type in AttributeValue.TYPE_MAP.items():
    identifier = AttributeValue.TYPE_URI_MAP[typeName]

    className = typeName + AttributeValue.CLASS_NAME_SUFFIX               
    classVars = {'TYPE': _type, 'IDENTIFIER': identifier}
    
    attributeValueClass = type(className, (AttributeValue, ), classVars)
    AttributeValue.register(attributeValueClass)
    _IDENTIFIER2CLASS_MAP[identifier] = attributeValueClass
    
    
class AttributeValueClassFactory(object):
    """Create AttributeValue types based on the XML namespace identifier
    
    Convenience wrapper for _IDENTIFIER2CLASS_MAP instance of 
    AttributeValueClassMap
    
    @ivar __classMap: mapping object to map attribute value URIs to their 
    implementations as classes
    @type __classMap: ndg.xacml.core.attributevalue.AttributeValueClassMap
    """
    __slots__ = ('__classMap',)
    
    def __init__(self, classMap=None):
        """Set a mapping object to map attribute value URIs to their 
        implementations as classes
        
        @param classMap: input an alternative to the default class mapping 
        object _IDENTIFIER2CLASS_MAP, if None, it will default to this setting
        @type classMap: ndg.xacml.core.attributevalue.AttributeValueClassMap
        """
        if classMap is None:
            self.__classMap = _IDENTIFIER2CLASS_MAP
        elif isinstance(classMap, AttributeValueClassMap):
            self.__classMap = classMap
        else:
            raise TypeError('Expecting %r derived type for "map" input; got %r'
                            % (AttributeValueClassMap, type(map)))
            
    def __call__(self, identifier):
        """Return <type>AttributeValue class for given identifier URI or None
        if no match is found
        
        @return: attribute value class
        @rtype: NoneType / ndg.xacml.core.attributevalue.AttributeValue derived
        type
        """
        return self.__classMap.get(identifier)
        


"""NDG XACML Condition type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "15/04/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: rule_combining_alg.py 7108 2010-06-28 09:19:26Z pjkersha $"
from abc import ABCMeta, abstractmethod

from ndg.xacml.core.context.result import Decision


# Rule Combining algorithms from the XACML spec.
ALGORITHMS = (
'urn:oasis:names:tc:xacml:1.0:rule-combining-algorithm:deny-overrides',
'urn:oasis:names:tc:xacml:1.0:policy-combining-algorithm:deny-overrides',
'urn:oasis:names:tc:xacml:1.0:rule-combining-algorithm:permit-overrides',
'urn:oasis:names:tc:xacml:1.0:policy-combining-algorithm:permit-overrides',
'urn:oasis:names:tc:xacml:1.0:rule-combining-algorithm:first-applicable',
'urn:oasis:names:tc:xacml:1.0:policy-combining-algorithm:first-applicable',
'urn:oasis:names:tc:xacml:1.0:policy-combining-algorithm:only-one-applicable',
'urn:oasis:names:tc:xacml:1.1:rule-combining-algorithm:ordered-deny-overrides',
'urn:oasis:names:tc:xacml:1.1:policy-combining-algorithm:ordered-deny-overrides',
'urn:oasis:names:tc:xacml:1.1:rule-combining-algorithm:ordered-permit-overrides',
'urn:oasis:names:tc:xacml:1.1:policy-combining-algorithm:ordered-permit-overrides',
)


class RuleCombiningAlgInterface(object):
    """Interface class for XAML rule combining algorithms"""
    
    @abstractmethod
    def evaluate(self, rules, context):
        """Combine the input rule results to make an access control decision 
        based.  Derived classes must implement this method.  This implementation
        returns indeterminate result.
        
        @param rules: rules from the policy.  Decisions from these will be put
        together into a single decision by this algorithm
        @type rules: TypedList(<ndg.xacml.core.rule.Rule>)
        @param context: request context to apply to the rules
        @type context: ndg.xacml.core.request.Request
        @return: resulting overall access control decision
        @rtype: ndg.xacml.core.context.result.Decision
        """
        return Decision.INDETERMINATE


class DenyOverridesRuleCombiningAlg(RuleCombiningAlgInterface):
    """Deny overrides rule combining algorithm"""
    
    def evaluate(self, rules, context):
        """Combine the input rule results to make an access control decision.
        Implementation taken direct from XACML 2.0 spec. pseudo code - Section
        C.1 Deny Overrides
        
        @param rules: rules from the policy.  Decisions from these will be put
        together into a single decision by this algorithm
        @type rules: TypedList(<ndg.xacml.core.rule.Rule>)
        @param context: request context to apply to the rules
        @type context: ndg.xacml.core.request.Request
        @return: resulting overall access control decision
        @rtype: ndg.xacml.core.context.result.Decision
        """
        atLeastOneError = False
        potentialDeny = False
        atLeastOnePermit = False
        
        for rule in rules:
            decision = rule.evaluate(context)
            if decision == Decision.DENY:
                return Decision.DENY

            if decision == Decision.PERMIT:
                atLeastOnePermit = True
                continue
            
            if decision == Decision.NOT_APPLICABLE:
                continue
            
            if decision == Decision.INDETERMINATE:
                atLeastOneError = True
    
                if effect(rule) == Decision.DENY:
                    potentialDeny = True
                    
                continue

        if potentialDeny:
            return Decision.INDETERMINATE

        elif atLeastOnePermit:
            return Decision.PERMIT
        
        elif atLeastOneError:
            return Decision.INDETERMINATE
        else:
            return Decision.NOT_APPLICABLE


class PermitOverridesRuleCombiningAlg(RuleCombiningAlgInterface):
    """Implementation of permit overrides XACML rule combining algorithm"""
    
    def evaluate(self, rules, context):
        """Combine the input rule results to make an access control decision.
        Implementation taken direct from XACML 2.0 spec. pseudo code - Section
        C.3
        
        @param rules: rules from the policy.  Decisions from these will be put
        together into a single decision by this algorithm
        @type rules: TypedList(<ndg.xacml.core.rule.Rule>)
        @param context: request context to apply to the rules
        @type context: ndg.xacml.core.request.Request
        @return: resulting overall access control decision
        @rtype: ndg.xacml.core.context.result.Decision
        """
        atLeastOneError = False
        potentialPermit = False
        atLeastOneDeny = False
        
        for rule in rules:
            decision = rule.evaluate(context)
            if decision == Decision.DENY:
                atLeastOneDeny = True
                continue
            
            if decision == Decision.PERMIT:
                return Decision.PERMIT
            
            if decision == Decision.NOT_APPLICABLE:
                continue
            
            if decision == Decision.INDETERMINATE:
                atLeastOneError = True
                
                if rule.effect.value == Decision.PERMIT_STR:
                    potentialPermit = True
                
                continue
        
        if potentialPermit:
            return Decision.INDETERMINATE
        
        if atLeastOneDeny:
            return Decision.DENY
        
        if atLeastOneError:
            return Decision.INDETERMINATE
        
        return Decision.NOT_APPLICABLE

    
class RuleCombiningAlgClassFactory(object):
    """Class Factory mapping Rule Combining Algorithm identifiers to their
    class implementations"""
    
    # All algorithms are not implemented by default(!)
    DEFAULT_MAP = {}.fromkeys(ALGORITHMS, NotImplemented)
    
    # Permit overrides is the only one currently implemented
    DEFAULT_MAP.update({
    'urn:oasis:names:tc:xacml:1.0:rule-combining-algorithm:deny-overrides':
        DenyOverridesRuleCombiningAlg,      
    'urn:oasis:names:tc:xacml:1.0:rule-combining-algorithm:permit-overrides':
        PermitOverridesRuleCombiningAlg
    })
    __slots__ = ('__map',)
    
    def __init__(self, map=DEFAULT_MAP):
        """Initialise mapping of identifiers to class implementations
        
        @param map: mapping of rule combining algorithms IDs to classes.  Set 
        this to override the default taken from the DEFAULT_MAP class variable
        """
        self.__map = map
    
    def __call__(self, identifier):
        """Return the class for a given Rule Combining Algorithm identifier
        @param identifier: XACML rule combining algorithm urn
        @type identifier: basestring
        @return: rule combining class corresponding to the given input
        identifier
        @rtype: RuleCombiningAlgInterface derived type or NoneType if no match 
        is found or NotImplementedType if the identifier corresponds to a valid 
        XACML rule combining algorithm but is not supported in this 
        implementation
        """
        return self.__map.get(identifier)
